'''
Functions for python 2/3 compatibility.
'''

from __future__ import absolute_import, division, print_function, unicode_literals

import sys

IS_PYTHON3 = sys.version_info >= (3,)

def typecode(t):
    '''Typecode handling for array module.

    Python 3 expects a unicode character, whereas python 2 expects a byte char.

    Arguments:

        `t` (typecode string):

            An input for array.array.

    Return value:

        The input formatted for the appropriate python version.
    '''
    if IS_PYTHON3:
        return t
    else:
        return chr(ord(t))

if IS_PYTHON3:
    def is_string(s):
        return isinstance(s, (str, bytes))
else:
    def is_string(s):
        return isinstance(s, basestring)

# array.tostring is deprecated in python3
if IS_PYTHON3:
    tobytes = lambda array: array.tobytes()
    frombytes = lambda array, src: array.frombytes(src)
else:
    tobytes = lambda array: array.tostring()
    frombytes = lambda array, src: array.fromstring(src)



'''
Runs unit tests of functions associated with spectral databases.

To run the unit tests, type the following from the system command line:

    # python -m spectral.tests.database

Note that the ECOSTRESS database must be requested so if the data files are
not located on the local file system, these tests will be skipped.
'''

from __future__ import absolute_import, division, print_function, unicode_literals

import numpy as np
import os
from numpy.testing import assert_almost_equal

import spectral as spy
from spectral.io.aviris import read_aviris_bands
from spectral.tests import testdir
from spectral.tests.spytest import SpyTest

ECOSTRESS_DATA_DIR = os.path.join(os.path.split(__file__)[0],
                                  'data/ecostress')
ECOSTRESS_DB = os.path.join(testdir, 'ecostress.db')
USGS_DATA_DIR = os.path.join(os.path.split(__file__)[0],
                             'data/usgs/ASCIIdata')
USGS_DB = os.path.join(testdir, 'usgs.db')
AVIRIS_BAND_FILE = os.path.join(os.path.split(__file__)[0], 'data/92AV3C.spc')


class ECOSTRESSDatabaseCreationTest(SpyTest):
    '''Tests ECOSTRESS database creation from text files.'''

    def __init__(self):
        pass

    def setup(self):
        if not os.path.isdir(testdir):
            os.makedirs(testdir)
        if os.path.exists(ECOSTRESS_DB):
            os.remove(ECOSTRESS_DB)

    def test_create_database(self):
        '''Test creating new database from ECOSTRESS data files.'''
        db = spy.EcostressDatabase.create(ECOSTRESS_DB,
                                          ECOSTRESS_DATA_DIR)
        assert(list(db.query('SELECT COUNT() FROM Spectra'))[0][0] == 3)


class ECOSTRESSDatabaseTest(SpyTest):
    '''Tests that ECOSTRESS database works properly'''

    def __init__(self):
        pass

    def setup(self):
        self.db = spy.EcostressDatabase(ECOSTRESS_DB)

    def test_read_signatures(self):
        '''Can get spectra from the opened database.'''
        assert(list(self.db.query('SELECT COUNT() FROM Spectra'))[0][0] == 3)

    def test_create_envi_lib(self):
        '''Can resample spectra and create an ENVI spectral library.'''
        bands = read_aviris_bands(AVIRIS_BAND_FILE)
        cursor = self.db.query('SELECT SpectrumID FROM Spectra')
        ids = [r[0] for r in cursor]
        bands.centers = [x / 1000. for x in bands.centers]
        bands.bandwidths = [x / 1000. for x in bands.bandwidths]
        slib = self.db.create_envi_spectral_library(ids, bands)
        assert(slib.spectra.shape == (3, 220))


class USGSDatabaseCreationTest(SpyTest):
    '''Tests USGS database creation from text files.'''

    def __init__(self):
        pass

    def setup(self):
        if not os.path.isdir(testdir):
            os.makedirs(testdir)
        if os.path.exists(USGS_DB):
            os.remove(USGS_DB)

    def test_create_database(self):
        '''Test creating new database from USGS data files.'''
        db = spy.USGSDatabase.create(USGS_DB, USGS_DATA_DIR)
        assert(list(db.query('SELECT COUNT() FROM Samples'))[0][0] == 8)
        assert(list(db.query('SELECT COUNT() FROM SpectrometerData'))
               [0][0] == 13)


class USGSDatabaseTest(SpyTest):
    '''Tests that USGS database works properly'''

    def __init__(self):
        pass

    def setup(self):
        self.db = spy.USGSDatabase(USGS_DB)

    def test_read_signatures(self):
        '''Can get spectra from the opened database.'''
        assert(list(self.db.query('SELECT COUNT() FROM Samples'))[0][0] == 8)
        assert(list(self.db.query('SELECT COUNT() FROM SpectrometerData'))
               [0][0] == 13)

        some_sample = list(self.db.query('''SELECT Chapter, FileName,
                    AssumedWLSpmeterDataID,
                    NumValues, MinValue, MaxValue
                    FROM Samples
                    WHERE LibName='liba' AND Record=1 AND
                    Description='Material a b0 0 ASDFRa AREF' AND
                    Spectrometer='ASDFR' AND Purity='a' AND MeasurementType='AREF'
                    '''))[0]
        assert(some_sample[0] == 'ChapterB_b0')
        assert(some_sample[1] == 'liba_Material_a_b0_0_ASDFRa_AREF.txt')
        assert(some_sample[3] == 24)
        assert_almost_equal(some_sample[4], 0.33387077)
        assert_almost_equal(some_sample[5], 0.51682192)

        some_spectrometer_data = list(self.db.query('''SELECT LibName, Record, MeasurementType, Unit,
                Name, Description, FileName, NumValues, MinValue, MaxValue
                FROM SpectrometerData
                WHERE SpectrometerDataID=?
                ''', (some_sample[2],)))[0]
        assert(some_spectrometer_data[0] == 'liba')
        assert(some_spectrometer_data[1] == 13)
        assert(some_spectrometer_data[2] == 'Wavelengths')
        assert(some_spectrometer_data[3] == 'micrometer')
        assert(some_spectrometer_data[4] == 'ASD')
        assert(some_spectrometer_data[5] == 'Wavelengths ASD 0.35-2.5 um')
        assert(some_spectrometer_data[6] ==
               'liba_Wavelengths_ASD_0.35-2.5_um.txt')
        assert(some_spectrometer_data[7] == 24)
        assert_almost_equal(some_spectrometer_data[8], 0.35)
        assert_almost_equal(some_spectrometer_data[9], 2.5)

    def test_get_spectrum(self):
        some_sample_id = list(self.db.query('''SELECT SampleID
            FROM Samples
            WHERE LibName='libc' AND Description='Material D 2 AVIRISb RTGC'
            '''))[0][0]
        (x, y) = self.db.get_spectrum(some_sample_id)
        assert(len(x) == len(y))
        assert(len(y) == 7)
        assert_almost_equal(y[0], 0.010381651)
        assert_almost_equal(x[-1], 2.2020326)

    def test_create_envi_lib(self):
        '''Can resample spectra and create an ENVI spectral library.'''
        bands = read_aviris_bands(AVIRIS_BAND_FILE)
        cursor = self.db.query('SELECT SampleID FROM Samples')
        ids = [r[0] for r in cursor]
        bands.centers = [x / 1000. for x in bands.centers]
        bands.bandwidths = [x / 1000. for x in bands.bandwidths]
        slib = self.db.create_envi_spectral_library(ids, bands)
        assert(slib.spectra.shape == (8, 220))


def run():
    print('\n' + '-' * 72)
    print('Running database tests.')
    print('-' * 72)
    for T in [ECOSTRESSDatabaseCreationTest, ECOSTRESSDatabaseTest, USGSDatabaseCreationTest, USGSDatabaseTest]:
        T().run()


if __name__ == '__main__':
    from spectral.tests.run import parse_args, reset_stats, print_summary
    import logging
    logging.getLogger('spectral').setLevel(logging.ERROR)
    parse_args()
    reset_stats()
    run()
    print_summary()



'''
Basic algorithms and data handling code.
'''

from __future__ import absolute_import, division, print_function, unicode_literals

import math
from numbers import Integral
import numpy as np
import pickle

import spectral as spy
from ..io.spyfile import SpyFile, TransformedImage
from ..utilities.errors import has_nan, NaNValueError
from .spymath import matrix_sqrt
from .transforms import LinearTransform

class Iterator:
    '''
    Base class for iterators over pixels (spectra).
    '''
    def __init__(self):
        pass

    def __iter__(self):
        raise NotImplementedError('Must override __iter__ in child class.')

    def get_num_elements(self):
        raise NotImplementedError(
            'Must override get_num_elements in child class.')

    def get_num_bands(self):
        raise NotImplementedError(
            'Must override get_num_bands in child class.')


class ImageIterator(Iterator):
    '''
    An iterator over all pixels in an image.
    '''
    def __init__(self, im):
        self.image = im
        self.numElements = im.shape[0] * im.shape[1]

    def get_num_elements(self):
        return self.numElements

    def get_num_bands(self):
        return self.image.shape[2]

    def __iter__(self):
        (M, N) = self.image.shape[:2]
        count = 0
        for i in range(M):
            self.row = i
            for j in range(N):
                self.col = j
                yield self.image[i, j]


class ImageMaskIterator(Iterator):
    '''
    An iterator over all pixels in an image corresponding to a specified mask.
    '''
    def __init__(self, image, mask, index=None):
        if mask.shape != image.shape[:len(mask.shape)]:
            raise ValueError('Mask shape does not match image.')
        self.image = image
        self.index = index
        # Get the proper mask for the training set
        if index:
            self.mask = np.equal(mask, index)
        else:
            self.mask = np.not_equal(mask, 0)
        self.n_elements = sum(self.mask.ravel())

    def get_num_elements(self):
        return self.n_elements

    def get_num_bands(self):
        return self.image.shape[2]

    def __iter__(self):
        coords = np.argwhere(self.mask)
        for (i, j) in coords:
            (self.row, self.col) = (i, j)
            yield self.image[i, j].astype(self.image.dtype).squeeze()

def iterator(image, mask=None, index=None):
    '''
    Returns an iterator over pixels in the image.

    Arguments:

        `image` (ndarray or :class:`spectral.Image`):

            An image over whose pixels will be iterated.

        `mask` (ndarray) [default None]:

            An array of integers that specify over which pixels in `image`
            iteration should be performed.

        `index` (int) [default None]:

            Specifies which value in `mask` should be used for iteration.

    Returns (:class:`spectral.Iterator`):

        An iterator over image pixels.

    If neither `mask` nor `index` are defined, iteration is performed over all
    pixels.  If `mask` (but not `index`) is defined, iteration is performed
    over all pixels for which `mask` is nonzero.  If both `mask` and `index`
    are defined, iteration is performed over all pixels `image[i,j]` for which
    `mask[i,j] == index`.
    '''

    if isinstance(image, Iterator):
        return image
    elif mask is not None:
        return ImageMaskIterator(image, mask, index)
    else:
        return ImageIterator(image)


def iterator_ij(mask, index=None):
    '''
    Returns an iterator over image pixel coordinates for a given mask.

    Arguments:

        `mask` (ndarray) [default None]:

            An array of integers that specify which coordinates should
            be returned.

        `index` (int) [default None]:

            Specifies which value in `mask` should be used for iteration.

    Returns:

        An iterator over image pixel coordinates. Each returned item is a
        2-tuple of the form (row, col).

    If `index` is not defined, iteration is performed over all non-zero
    elements.  If `index` is defined, iteration is performed over all
    coordinates for whch `mask[i,j] == index`.
    '''

    if mask.ndim != 2:
        raise ValueError('Invalid mask shape.')
    if index is None:
        mask = mask != 0
    else:
        mask = mask == index
    for rc in np.argwhere(mask):
        yield tuple(rc)


def mean_cov(image, mask=None, index=None):
    '''
    Return the mean and covariance of the set of vectors.

    Usage::

        (mean, cov, S) = mean_cov(vectors [, mask=None [, index=None]])

    Arguments:

        `image` (ndarrray, :class:`~spectral.Image`, or :class:`spectral.Iterator`):

            If an ndarray, it should have shape `MxNxB` and the mean &
            covariance will be calculated for each band (third dimension).

        `mask` (ndarray):

            If `mask` is specified, mean & covariance will be calculated for
            all pixels indicated in the mask array.  If `index` is specified,
            all pixels in `image` for which `mask == index` will be used;
            otherwise, all nonzero elements of `mask` will be used.

        `index` (int):

            Specifies which value in `mask` to use to select pixels from
            `image`. If not specified but `mask` is, then all nonzero elements
            of `mask` will be used.

        If neither `mask` nor `index` are specified, all samples in `vectors`
        will be used.

    Returns a 3-tuple containing:

        `mean` (ndarray):

            The length-`B` mean vectors

        `cov` (ndarray):

            The `BxB` unbiased estimate (dividing by N-1) of the covariance
            of the vectors.

        `S` (int):

            Number of samples used to calculate mean & cov

    Calculate the mean and covariance of of the given vectors. The argument
    can be an Iterator, a SpyFile object, or an `MxNxB` array.
    '''
    status = spy._status

    if isinstance(image, np.ndarray):
        X = image.astype(np.float64)
        if X.ndim == 3:
            X = image.reshape(-1, image.shape[-1]).T
        if mask is not None:
            mask = mask.ravel()
            if index is not None:
                ii = np.argwhere(mask == index)
            else:
                ii = np.argwhere(mask != 0)
            X = np.take(X, ii.squeeze(), axis=1)
        m = np.average(X, axis=1)
        C = np.cov(X)
        return (m, C, X.shape[1])

    if not isinstance(image, Iterator):
        it = iterator(image, mask, index)
    else:
        it = image

    nSamples = it.get_num_elements()
    B = it.get_num_bands()

    sumX = np.zeros((B,), 'd')
    sumX2 = np.zeros((B, B), 'd')
    count = 0

    statusInterval = max(1, nSamples / 100)
    status.display_percentage('Covariance.....')
    for x in it:
        if not count % statusInterval:
            status.update_percentage(float(count) / nSamples * 100.)
        count += 1
        sumX += x
        x = x.astype(np.float64)[:, np.newaxis]
        sumX2 += x.dot(x.T)
    mean = (sumX / count)
    sumX = sumX[:, np.newaxis]
    cov = (sumX2 - sumX.dot(sumX.T) / count) / (count - 1)
    status.end_percentage()
    return (mean, cov, count)


def cov_avg(image, mask, weighted=True):
    '''Calculates the covariance averaged over a set of classes.

    Arguments:

        `image` (ndarrray, :class:`~spectral.Image`, or :class:`spectral.Iterator`):

            If an ndarray, it should have shape `MxNxB` and the mean &
            covariance will be calculated for each band (third dimension).

        `mask` (integer-valued ndarray):

            Elements specify the classes associated with pixels in `image`.
            All pixels associeted with non-zero elements of `mask` will be
            used in the covariance calculation.

        `weighted` (bool, default True):

            Specifies whether the individual class covariances should be
            weighted when computing the average. If True, each class will
            be weighted by the number of pixels provided for the class;
            otherwise, a simple average of the class covariances is performed.

    Returns a class-averaged covariance matrix. The number of covariances used
    in the average is equal to the number of non-zero elements of `mask`.
    '''
    ids = set(mask.ravel()) - set((0,))
    classes = [calc_stats(image, mask, i) for i in ids]
    N = sum([c.nsamples for c in classes])
    if weighted:
        return np.sum([((c.nsamples - 1) / float(N - 1)) * c.cov
                       for c in classes], axis=0, dtype=np.float64)
    else:
        return np.mean([c.cov for c in classes], axis=0, dtype=np.float64)

def covariance(*args):
    '''
    Returns the covariance of the set of vectors.

    Usage::

        C = covariance(vectors [, mask=None [, index=None]])

    Arguments:

        `vectors` (ndarrray, :class:`~spectral.Image`, or :class:`spectral.Iterator`):

            If an ndarray, it should have shape `MxNxB` and the mean &
            covariance will be calculated for each band (third dimension).

        `mask` (ndarray):

            If `mask` is specified, mean & covariance will be calculated for
            all pixels indicated in the mask array.  If `index` is specified,
            all pixels in `image` for which `mask == index` will be used;
            otherwise, all nonzero elements of `mask` will be used.

        `index` (int):

            Specifies which value in `mask` to use to select pixels from
            `image`. If not specified but `mask` is, then all nonzero elements
            of `mask` will be used.

        If neither `mask` nor `index` are specified, all samples in `vectors`
        will be used.

    Returns:

        `C` (ndarray):

            The `BxB` unbiased estimate (dividing by N-1) of the covariance
            of the vectors.


    To also return the mean vector and number of samples, call
    :func:`~spectral.algorithms.algorithms.mean_cov` instead.
    '''
    return mean_cov(*args)[1]


class PrincipalComponents:
    '''
    An object for storing a data set's principal components.  The
    object has the following members:

        `eigenvalues`:

            A length B array of eigenvalues sorted in descending order

        `eigenvectors`:

            A `BxB` array of normalized eigenvectors (in columns)

        `stats` (:class:`GaussianStats`):

            A statistics object containing `mean`, `cov`, and `nsamples`.

        `transform`:

            A callable function to transform data to the space of the
            principal components.

        `reduce`:

            A method to return a reduced set of principal components based
            on either a fixed number of components or a fraction of total
            variance.

        `denoise`:

            A callable function to denoise data using a reduced set of
            principal components.

        `get_denoising_transform`:

            A callable function that returns a function for denoising data.
    '''
    def __init__(self, vals, vecs, stats):
        self.eigenvalues = vals
        self.eigenvectors = vecs
        self.stats = stats
        self.transform = LinearTransform(self.eigenvectors.T, pre=-self.mean)

    @property
    def mean(self):
        return self.stats.mean

    @property
    def cov(self):
        return self.stats.cov

    def reduce(self, N=0, **kwargs):
        '''Reduces the number of principal components.

        Keyword Arguments (one of the following must be specified):

            `num` (integer):

                Number of eigenvalues/eigenvectors to retain.  The top `num`
                eigenvalues will be retained.

            `eigs` (list):

                A list of indices of eigenvalues/eigenvectors to be retained.

            `fraction` (float):

                The fraction of total image variance to retain.  Eigenvalues
                will be retained (starting from greatest to smallest) until
                `fraction` of total image variance is retained.
        '''
        status = spy._status

        num = kwargs.get('num', None)
        eigs = kwargs.get('eigs', None)
        fraction = kwargs.get('fraction', None)
        if num is not None:
            return PrincipalComponents(self.eigenvalues[:num],
                                       self.eigenvectors[:, :num],
                                       self.stats)
        elif eigs is not None:
            vals = self.eigenvalues[eigs]
            vecs = self.eigenvectors[:, eigs]
            return PrincipalComponents(vals, vecs, self.stats)
        elif fraction is not None:
            if not 0 < fraction <= 1:
                raise Exception('fraction must be in range (0,1].')
            N = len(self.eigenvalues)
            cumsum = np.cumsum(self.eigenvalues)
            sum = cumsum[-1]
            # Count how many values to retain.
            for i in range(N):
                if (cumsum[i] / sum) >= fraction:
                    break
            if i == (N - 1):
                # No reduction
                status.write('No reduction in eigenvectors achieved.')
                return self

            vals = self.eigenvalues[:i + 1]
            vecs = self.eigenvectors[:, :i + 1]
            return PrincipalComponents(vals, vecs, self.stats)
        else:
            raise Exception('Must specify one of the following keywords:'
                            '`num`, `eigs`, `fraction`.')

    def denoise(self, X, **kwargs):
        '''Returns a de-noised version of `X`.

        Arguments:

            `X` (np.ndarray):

                Data to be de-noised. Can be a single pixel or an image.

        Keyword Arguments (one of the following must be specified):

            `num` (integer):

                Number of eigenvalues/eigenvectors to use.  The top `num`
                eigenvalues will be used.

            `eigs` (list):

                A list of indices of eigenvalues/eigenvectors to be used.

            `fraction` (float):

                The fraction of total image variance to retain.  Eigenvalues
                will be included (starting from greatest to smallest) until
                `fraction` of total image variance is retained.

        Returns denoised image data with same shape as `X`.

        Note that calling this method is equivalent to calling the
        `get_denoising_transform` method with same keyword and applying the
        returned transform to `X`. If you only intend to denoise data with the
        same parameters multiple times, then it is more efficient to get the
        denoising transform and reuse it, rather than calling this method
        multilple times.
        '''
        f = self.get_denoising_transform(**kwargs)
        return f(X)

    def get_denoising_transform(self, **kwargs):
        '''Returns a function for denoising image data.

        Keyword Arguments (one of the following must be specified):

            `num` (integer):

                Number of eigenvalues/eigenvectors to use.  The top `num`
                eigenvalues will be used.

            `eigs` (list):

                A list of indices of eigenvalues/eigenvectors to be used.

            `fraction` (float):

                The fraction of total image variance to retain.  Eigenvalues
                will be included (starting from greatest to smallest) until
                `fraction` of total image variance is retained.

        Returns a callable :class:`~spectral.algorithms.transforms.LinearTransform`
        object for denoising image data.
        '''
        V = self.reduce(self, **kwargs).eigenvectors
        f = LinearTransform(V.dot(V.T), pre=-self.mean,
                            post=self.mean)
        return f


def principal_components(image):
    '''
    Calculate Principal Component eigenvalues & eigenvectors for an image.

    Usage::

        pc = principal_components(image)

    Arguments:

        `image` (ndarray, :class:`spectral.Image`, :class:`GaussianStats`):

            An `MxNxB` image

    Returns a :class:`~spectral.algorithms.algorithms.PrincipalComponents`
    object with the following members:

        `eigenvalues`:

            A length B array of eigenvalues

        `eigenvectors`:

            A `BxB` array of normalized eigenvectors

        `stats` (:class:`GaussianStats`):

            A statistics object containing `mean`, `cov`, and `nsamples`.

        `transform`:

            A callable function to transform data to the space of the
            principal components.

        `reduce`:

            A method to reduce the number of eigenvalues.

        `denoise`:

            A callable function to denoise data using a reduced set of
            principal components.

        `get_denoising_transform`:

            A callable function that returns a function for denoising data.
    '''
    if isinstance(image, GaussianStats):
        stats = image
    else:
        stats = calc_stats(image)

    (L, V) = np.linalg.eig(stats.cov)

    # numpy says eigenvalues may not be sorted so we'll sort them, if needed.
    if not np.alltrue(np.diff(L) <= 0):
        ii = list(reversed(np.argsort(L)))
        L = L[ii]
        V = V[:, ii]

    return PrincipalComponents(L, V, stats)


class FisherLinearDiscriminant:
    '''
    An object for storing a data set's linear discriminant data.  For `C`
    classes with `B`-dimensional data, the object has the following members:

        `eigenvalues`:

            A length `C-1` array of eigenvalues

        `eigenvectors`:

            A `BxC` array of normalized eigenvectors

        `mean`:

            The length `B` mean vector of the image pixels (from all classes)

        `cov_b`:

            The `BxB` matrix of covariance *between* classes

        `cov_w`:

            The `BxB` matrix of average covariance *within* each class

        `transform`:

            A callable function to transform data to the space of the
            linear discriminant.
    '''
    def __init__(self, vals, vecs, mean, cov_b, cov_w):
        self.eigenvalues = vals
        self.eigenvectors = vecs
        self.mean = mean
        self.cov_b = cov_b
        self.cov_w = cov_w
        self.transform = LinearTransform(self.eigenvectors.T, pre=-self.mean)


def linear_discriminant(classes, whiten=True):
    '''
    Solve Fisher's linear discriminant for eigenvalues and eigenvectors.

    Usage: (L, V, Cb, Cw) = linear_discriminant(classes)

    Arguments:

        `classes` (:class:`~spectral.algorithms.TrainingClassSet`):

            The set of `C` classes to discriminate.

    Returns a `FisherLinearDiscriminant` object containing the within/between-
    class covariances, mean vector, and a callable transform to convert data to
    the transform's space.

    This function determines the solution to the generalized eigenvalue problem

            Cb * x = lambda * Cw * x

    Since cov_w is normally invertable, the reduces to

            (inv(Cw) * Cb) * x = lambda * x

    References:

        Richards, J.A. & Jia, X. Remote Sensing Digital Image Analysis: An
        Introduction. (Springer: Berlin, 1999).
    '''
    C = len(classes)            # Number of training sets
    rank = len(classes) - 1

    classes.calc_stats()

    # Calculate total # of training pixels and total mean
    N = 0
    B = classes.nbands
    K = len(classes)
    mean = np.zeros(B, dtype=np.float64)
    for s in classes:
        N += s.size()
        mean += s.size() * s.stats.mean
    mean /= N

    cov_b = np.zeros((B, B), np.float64)            # cov between classes
    cov_w = np.zeros((B, B), np.float64)            # cov within classes
    for s in classes:
        cov_w += ((s.size() - 1) / float(N - 1)) * s.stats.cov
        m = s.stats.mean - mean
        cov_b += (s.size() / float(N) / (K - 1)) * np.outer(m, m)

    inv_cov_w = np.linalg.inv(cov_w)
    (vals, vecs) = np.linalg.eig(inv_cov_w.dot(cov_b))
    vals = vals[:rank]
    vecs = vecs[:, :rank]

    if whiten:
        # Diagonalize cov_within in the new space
        v = vecs.T.dot(cov_w).dot(vecs)
        d = np.sqrt(np.diag(v) * np.diag(v).conj())
        for i in range(vecs.shape[1]):
            vecs[:, i] /= math.sqrt(d[i].real)

    return FisherLinearDiscriminant(vals.real, vecs.real, mean, cov_b, cov_w)

# Alias for Linear Discriminant Analysis (LDA)
lda = linear_discriminant


def log_det(x):
    return sum(np.log([eigv for eigv in np.linalg.eigvals(x)
                          if eigv > 0]))


class GaussianStats(object):
    '''A class for storing Gaussian statistics for a data set.

    Statistics stored include:

        `mean`:

            Mean vector

        `cov`:

            Covariance matrix

        `nsamples`:

            Number of samples used in computing the statistics

    Several derived statistics are computed on-demand (and cached) and are
    available as property attributes. These include:

        `inv_cov`:

            Inverse of the covariance

        `sqrt_cov`:

            Matrix square root of covariance: sqrt_cov.dot(sqrt_cov) == cov

        `sqrt_inv_cov`:

            Matrix square root of the inverse of covariance

        `log_det_cov`:

            The log of the determinant of the covariance matrix

        `principal_components`:

            The principal components of the data, based on mean and cov.
    '''

    def __init__(self, mean=None, cov=None, nsamples=None, inv_cov=None):
        self.cov = cov
        self._inv_cov = inv_cov
        self.mean = mean
        self.nsamples = nsamples

    @property
    def cov(self):
        '''Property method returning the covariance matrix.'''
        return self._cov

    @cov.setter
    def cov(self, C):
        self.reset_derived_stats()
        self._cov = C

    @property
    def inv_cov(self):
        '''Property method returning the inverse of the covariance matrix.'''
        if self._inv_cov is None:
            self._inv_cov = np.linalg.inv(self._cov)
        return self._inv_cov

    def reset_derived_stats(self):
        self._cov = self._inv_cov = None
        self._sqrt_cov = self._sqrt_inv_cov = self._pcs = None
        self._log_det_cov = None

    @property
    def sqrt_cov(self):
        '''Property method returning the matrix square root of the covariance.
        If `C` is the covariance, then the returned value is a matrix `S`
        such that S.dot(S) == C.
        '''
        if self._sqrt_cov is None:
            pcs = self.principal_components
            self._sqrt_cov = matrix_sqrt(eigs=(pcs.eigenvalues,
                                               pcs.eigenvectors),
                                         symmetric=True)
        return self._sqrt_cov

    @property
    def sqrt_inv_cov(self):
        '''Property method returning matrix square root of inverse of cov.
        If `C` is the covariance, then the returned value is a matrix `S`
        such that S.dot(S) == inv(C).
        '''
        if self._sqrt_inv_cov is None:
            pcs = self.principal_components
            self._sqrt_inv_cov = matrix_sqrt(eigs=(pcs.eigenvalues,
                                                   pcs.eigenvectors),
                                             symmetric=True,
                                             inverse=True)
        return self._sqrt_inv_cov

    @property
    def principal_components(self):
        if self._pcs is None:
            (evals, evecs) = np.linalg.eigh(self._cov)
            self._pcs = PrincipalComponents(evals, evecs, self)
        return self._pcs

    @property
    def log_det_cov(self):
        if self._log_det_cov is None:
            evals = self.principal_components.eigenvalues
            self._log_det_cov = np.sum(np.log([v for v in evals if v > 0]))
        return self._log_det_cov

    def transform(self, xform):
        '''Returns a version of the stats transformed by a linear transform.'''
        if not isinstance(xform, LinearTransform):
            raise TypeError('Expected a LinearTransform object.')
        m = xform(self.mean)
        C = xform._A.dot(self.cov).dot(xform._A.T)
        return GaussianStats(mean=m, cov=C, nsamples=self.nsamples)

    def get_whitening_transform(self):
        '''Returns transform that centers and whitens data for these stats.'''
        C_1 = np.linalg.inv(self.cov)
        return LinearTransform(matrix_sqrt(C_1, True), pre=-self.mean)


def calc_stats(image, mask=None, index=None, allow_nan=False):
    '''Computes Gaussian stats for image data..

    Arguments:

        `image` (ndarrray, :class:`~spectral.Image`, or :class:`spectral.Iterator`):

            If an ndarray, it should have shape `MxNxB` and the mean &
            covariance will be calculated for each band (third dimension).

        `mask` (ndarray):

            If `mask` is specified, mean & covariance will be calculated for
            all pixels indicated in the mask array.  If `index` is specified,
            all pixels in `image` for which `mask == index` will be used;
            otherwise, all nonzero elements of `mask` will be used.

        `index` (int):

            Specifies which value in `mask` to use to select pixels from
            `image`. If not specified but `mask` is, then all nonzero elements
            of `mask` will be used.

        `allow_nan` (bool, default False):

            If True, statistics will be computed even if `np.nan` values are
            present in the data; otherwise, `~spectral.algorithms.spymath.NaNValueError`
            is raised.

        If neither `mask` nor `index` are specified, all samples in `vectors`
        will be used.

    Returns:

        `GaussianStats` object:

            This object will have members `mean`, `cov`, and `nsamples`.
    '''
    (mean, cov, N) = mean_cov(image, mask, index)
    if has_nan(mean) and not allow_nan:
        raise NaNValueError('NaN values present in data.')
    return GaussianStats(mean=mean, cov=cov, nsamples=N)


class TrainingClass:
    def __init__(self, image, mask, index=0, class_prob=1.0):
        '''Creates a new training class defined by applying `mask` to `image`.

        Arguments:

            `image` (:class:`spectral.Image` or :class:`numpy.ndarray`):

                The `MxNxB` image over which the training class is defined.

            `mask` (:class:`numpy.ndarray`):

                An `MxN` array of integers that specifies which pixels in
                `image` are associated with the class.

            `index` (int) [default 0]:

                if `index` == 0, all nonzero elements of `mask` are associated
                with the class.  If `index` is nonzero, all elements of `mask`
                equal to `index` are associated with the class.

            `class_prob` (float) [default 1.0]:

                Defines the prior probability associated with the class, which
                is used in maximum likelihood classification.  If `classProb`
                is 1.0, prior probabilities are ignored by classifiers, giving
                all class equal weighting.
        '''
        self.image = image
        if image is not None:
            self.nbands = image.shape[2]
        self.nbands = None
        self.mask = mask
        self.index = index
        self.class_prob = class_prob
        self.stats = None

        self._stats_valid = False

    def __iter__(self):
        '''Returns an iterator over all samples for the class.'''
        it = ImageMaskIterator(self.image, self.mask, self.index)
        for i in it:
            yield i

    def stats_valid(self, tf=None):
        '''
        Sets statistics for the TrainingClass to be valid or invalid.

        Arguments:

            `tf` (bool or None):

                A value evaluating to False indicates that statistics should be
                recalculated prior to being used. If the argument is `None`,
                a value will be returned indicating whether stats need to be
                recomputed.
        '''
        if tf is None:
            return self._stats_valid
        self._stats_valid = tf

    def size(self):
        '''Returns the number of pixels/samples in the training set.'''

        # If the stats are invalid, the number of pixels in the
        # training set may have changed.
        if self._stats_valid:
            return self.stats.nsamples

        if self.index:
            return np.sum(np.equal(self.mask, self.index).ravel())
        else:
            return np.sum(np.not_equal(self.mask, 0).ravel())


    def calc_stats(self):
        '''
        Calculates statistics for the class.

        This function causes the :attr:`stats` attribute of the class to be
        updated, where `stats` will have the following attributes:

        =============  ======================   ===================================
        Attribute      Type                          Description
        =============  ======================   ===================================
        `mean`         :class:`numpy.ndarray`   length-`B` mean vector
        `cov`          :class:`numpy.ndarray`   `BxB` covariance matrix
        `inv_cov`      :class:`numpy.ndarray`   Inverse of `cov`
        `log_det_cov`  float                    Natural log of determinant of `cov`
        =============  ======================   ===================================
        '''
        self.stats = calc_stats(self.image, self.mask, self.index)
        self.nbands = self.image.shape[-1]
        self._stats_valid = True

    def transform(self, transform):
        '''
        Perform a linear transformation on the statistics of the training set.

        Arguments:

            `transform` (:class:numpy.ndarray or LinearTransform):

                The linear transform array.  If the class has `B` bands, then
                `transform` must have shape `(C,B)`.

        After `transform` is applied, the class statistics will have `C` bands.
        '''
        if isinstance(transform, np.ndarray):
            transform = LinearTransform(transform)
        self.stats.mean = transform(self.stats.mean)
        self.stats.cov = np.dot(
            transform._A, self.stats.cov).dot(transform._A.T)
        self.nbands = transform.dim_out


class SampleIterator:
    '''Iterator over all classes and samples in a TrainingClassSet object.'''
    def __init__(self, trainingData):
        self.classes = trainingData

    def __iter__(self):
        for cl in self.classes:
            for sample in cl:
                yield sample


class TrainingClassSet:
    '''A class to manage a set of :class:`~spectral.TrainingClass` objects.'''
    def __init__(self):
        self.classes = {}
        self.nbands = None

    def __getitem__(self, i):
        '''Returns the training class having ID i.'''
        return self.classes[i]

    def __len__(self):
        '''Returns number of training classes in the set.'''
        return len(self.classes)

    def add_class(self, cl):
        '''Adds a new class to the training set.

        Arguments:

            `cl` (:class:`spectral.TrainingClass`):

                `cl.index` must not duplicate a class already in the set.
        '''
        if cl.index in self.classes:
            raise Exception('Attempting to add class with duplicate index.')
        self.classes[cl.index] = cl
        if not self.nbands:
            self.nbands = cl.nbands

    def transform(self, X):
        '''Applies linear transform, M, to all training classes.

        Arguments:

            `X` (:class:numpy.ndarray):

                The linear transform array.  If the classes have `B` bands, then
                `X` must have shape `(C,B)`.

        After the transform is applied, all classes will have `C` bands.
        '''
        for cl in list(self.classes.values()):
            cl.transform(X)
        self.nbands = list(self.classes.values())[0].nbands

    def __iter__(self):
        '''An iterator over all training classes in the set.'''
        for cl in list(self.classes.values()):
            yield cl

    def all_samples(self):
        '''An iterator over all samples in all classes.'''
        return SampleIterator(self)

    def calc_stats(self):
        '''Computes statistics for each class, if not already computed.'''
        for c in list(self.classes.values()):
            if not c.stats_valid():
                c.calc_stats()
        self.nbands = list(self.classes.values())[0].nbands

    def save(self, filename, calc_stats=False):
        for c in list(self.classes.values()):
            if c.stats is None:
                if calc_stats == False:
                    msg = 'Class statistics are missing from at least one ' \
                      'class and are required to save the training class ' \
                      'data. Call the `save` method with keyword ' \
                      '`calc_stats=True` if you want to compute them and ' \
                      'then save the class data.'
                    raise Exception (msg)
                else:
                    c.calc_stats()
        f = open(filename, 'wb')
        ids = sorted(self.classes.keys())
        pickle.dump(self.classes[ids[0]].mask, f)
        pickle.dump(len(self), f)
        for id in ids:
            c = self.classes[id]
            pickle.dump(c.index, f)
            pickle.dump(c.stats.cov, f)
            pickle.dump(c.stats.mean, f)
            pickle.dump(c.stats.nsamples, f)
            pickle.dump(c.class_prob, f)
        f.close()

    def load(self, filename, image):
        f = open(filename, 'rb')
        mask = pickle.load(f)
        nclasses = pickle.load(f)
        for i in range(nclasses):
            index = pickle.load(f)
            cov = pickle.load(f)
            mean = pickle.load(f)
            nsamples = pickle.load(f)
            class_prob = pickle.load(f)
            c = TrainingClass(image, mask, index, class_prob)
            c.stats = GaussianStats(mean=mean, cov=cov, nsamples=nsamples)
            if not (cov is None or mean is None or nsamples is None):
                c.stats_valid(True)
                c.nbands = len(mean)
            self.add_class(c)
        f.close


def create_training_classes(image, class_mask, calc_stats=False, indices=None):
    '''
    Creates a :class:spectral.algorithms.TrainingClassSet: from an indexed array.

    USAGE:  sets = createTrainingClasses(classMask)

    Arguments:

        `image` (:class:`spectral.Image` or :class:`numpy.ndarray`):

            The image data for which the training classes will be defined.
            `image` has shape `MxNxB`.

        `class_mask` (:class:`numpy.ndarray`):

            A rank-2 array whose elements are indices of various spectral
            classes.  if `class_mask[i,j]` == `k`, then `image[i,j]` is
            assumed to belong to class `k`.

        `calc_stats` (bool):

            An optional parameter which, if True, causes statistics to be
            calculated for all training classes.

    Returns:

        A :class:`spectral.algorithms.TrainingClassSet` object.

    The dimensions of classMask should be the same as the first two dimensions
    of the corresponding image. Values of zero in classMask are considered
    unlabeled and are not added to a training set.
    '''

    if indices is not None:
        class_indices = set(indices) - set((0,))
    else:
        class_indices = set(class_mask.ravel()) - set((0,))
    classes = TrainingClassSet()
    classes.nbands = image.shape[-1]
    for i in class_indices:
        cl = TrainingClass(image, class_mask, i)
        if calc_stats:
            cl.calc_stats()
        classes.add_class(cl)
    return classes


def ndvi(data, red, nir):
    '''Calculates Normalized Difference Vegetation Index (NDVI).

    Arguments:

        `data` (ndarray or :class:`spectral.Image`):

            The array or SpyFile for which to calculate the index.

        `red` (int or int range):

            Index of the red band or an index range for multiple bands.

        `nir` (int or int range):

            An integer index of the near infrared band or an index range for
            multiple bands.

    Returns an ndarray:

        An array containing NDVI values in the range [-1.0, 1.0] for each
        corresponding element of data.
    '''

    r = data[:, :, red].astype(float)
    if len(r.shape) == 3 and r.shape[2] > 1:
        r = sum(r, 2) / r.shape[2]
    n = data[:, :, nir].astype(float)
    if len(n.shape) == 3 and n.shape[2] > 1:
        n = sum(n, 2) / n.shape[2]

    return (n - r) / (n + r)


def bdist(class1, class2):
    '''
    Calulates the Bhattacharyya distance between two classes.

    USAGE:  bd = bdist(class1, class2)

    Arguments:

        `class1`, `class2` (:class:`~spectral.algorithms.algorithms.TrainingClass`)

    Returns:

        A float value for the Bhattacharyya Distance between the classes.  This
        function is aliased to :func:`~spectral.algorithms.algorithms.bDistance`.

    References:

        Richards, J.A. & Jia, X. Remote Sensing Digital Image Analysis: An
        Introduction. (Springer: Berlin, 1999).
    '''
    terms = bdist_terms(class1, class2)
    return terms[0] + terms[1]

bDistance = bdist


def bdist_terms(a, b):
    '''
    Calulate the linear and quadratic terms of the Bhattacharyya distance
    between two classes.

    USAGE:  (linTerm, quadTerm) = bDistanceTerms(a, b)

    ARGUMENTS:
        (a, b)              The classes for which to determine the
                            B-distance.
    RETURN VALUE:
                            A 2-tuple of the linear and quadratic terms
    '''
    m = a.stats.mean - b.stats.mean
    avgCov = (a.stats.cov + b.stats.cov) / 2

    lin_term = (1 / 8.) * np.dot(np.transpose(m), np.dot(np.inv(avgCov), m))

    quad_term = 0.5 * (log_det(avgCov)
                       - 0.5 * a.stats.log_det_cov
                       - 0.5 * b.stats.log_det_cov)

    return (lin_term, float(quad_term))


def transform_image(matrix, image):
    '''
    Performs linear transformation on all pixels in an image.

    Arguments:

        matrix (:class:`numpy.ndarray`):

            A `CxB` linear transform to apply.

        image  (:class:`numpy.ndarray` or :class:`spectral.Image`):

            Image data to transform

    Returns:

        If `image` is an `MxNxB` :class:`numpy.ndarray`, the return will be a
        transformed :class:`numpy.ndarray` with shape `MxNxC`.  If `image` is
        :class:`spectral.Image`, the returned object will be a
        :class:`spectral.TransformedImage` object and no transformation of data
        will occur until elements of the object are accessed.
    '''
    if isinstance(image, SpyFile):
        return TransformedImage(matrix, image)
    elif isinstance(image, np.ndarray):
        (M, N, B) = image.shape
        ximage = np.zeros((M, N, matrix.shape[0]), float)

        for i in range(M):
            for j in range(N):
                ximage[i, j] = np.dot(matrix, image[i, j].astype(float))
        return ximage
    else:
        raise 'Unrecognized image type passed to transform_image.'


def orthogonalize(vecs, start=0):
    '''
    Performs Gram-Schmidt Orthogonalization on a set of vectors.

    Arguments:

        `vecs` (:class:`numpy.ndarray`):

            The set of vectors for which an orthonormal basis will be created.
            If there are `C` vectors of length `B`, `vecs` should be `CxB`.

        `start` (int) [default 0]:

            If `start` > 0, then `vecs[start]` will be assumed to already be
            orthonormal.

    Returns:

        A new `CxB` containing an orthonormal basis for the given vectors.
    '''
    (M, N) = vecs.shape
    basis = np.array(np.transpose(vecs))
    eye = np.identity(N).astype(float)
    for i in range(start, M):
        if i == 0:
            basis[:, 0] /= np.linalg.norm(basis[:, 0])
            continue
        v = basis[:, i] / np.linalg.norm(basis[:, i])
        U = basis[:, :i]
        P = eye - U.dot(np.linalg.inv(U.T.dot(U)).dot(U.T))
        basis[:, i] = P.dot(v)
        basis[:, i] /= np.linalg.norm(basis[:, i])
    return np.transpose(basis)


def unmix(data, members):
    '''
    Perform linear unmixing on image data.

    USAGE: mix = unmix(data, members)

    ARGUMENTS:
        data                The MxNxB image data to be unmixed
        members             An CxB array of C endmembers
    RETURN VALUE:
        mix                 An MxNxC array of endmember fractions.

    unmix performs linear unmixing on the image data.  After calling the
    function, mix[:,:,i] will then represent the fractional abundances
    for the i'th endmember. If the result of unmix is returned into 'mix',
    then an array of indices of greatest fractional endmembers is obtained
    by argmax(mix).

    Note that depending on endmembers given, fractional abundances for
    endmembers may be negative.
    '''
    assert members.shape[1] == data.shape[2], \
        'Matrix dimensions are not aligned.'

    members = members.astype(float)
    # Calculate the pseudo inverse
    pi = np.dot(members, np.transpose(members))
    pi = np.dot(np.inv(pi), members)

    (M, N, B) = data.shape
    unmixed = np.zeros((M, N, members.shape[0]), float)
    for i in range(M):
        for j in range(N):
            unmixed[i, j] = np.dot(pi, data[i, j].astype(float))
    return unmixed


def spectral_angles(data, members):
    '''Calculates spectral angles with respect to given set of spectra.

    Arguments:

        `data` (:class:`numpy.ndarray` or :class:`spectral.Image`):

            An `MxNxB` image for which spectral angles will be calculated.

        `members` (:class:`numpy.ndarray`):

            `CxB` array of spectral endmembers.

    Returns:

        `MxNxC` array of spectral angles.


    Calculates the spectral angles between each vector in data and each of the
    endmembers.  The output of this function (angles) can be used to classify
    the data by minimum spectral angle by calling argmin(angles).
    '''
    assert members.shape[1] == data.shape[2], \
        'Matrix dimensions are not aligned.'

    m = np.array(members, np.float64)
    m /= np.sqrt(np.einsum('ij,ij->i', m, m))[:, np.newaxis]

    norms = np.sqrt(np.einsum('ijk,ijk->ij', data, data))
    dots = np.einsum('ijk,mk->ijm', data, m)
    dots = np.clip(dots / norms[:, :, np.newaxis], -1, 1)
    return np.arccos(dots)

def msam(data, members):
    '''Modified SAM scores according to Oshigami, et al [1]. Endmembers are
    mean-subtracted prior to spectral angle calculation. Results are
    normalized such that the maximum value of 1 corresponds to a perfect match
    (zero spectral angle).

    Arguments:

        `data` (:class:`numpy.ndarray` or :class:`spectral.Image`):

            An `MxNxB` image for which spectral angles will be calculated.

        `members` (:class:`numpy.ndarray`):

            `CxB` array of spectral endmembers.

    Returns:

        `MxNxC` array of MSAM scores with maximum value of 1 corresponding
        to a perfect match (zero spectral angle).

    Calculates the spectral angles between each vector in data and each of the
    endmembers.  The output of this function (angles) can be used to classify
    the data by minimum spectral angle by calling argmax(angles).

    References:

    [1] Shoko Oshigami, Yasushi Yamaguchi, Tatsumi Uezato, Atsushi Momose,
    Yessy Arvelyna, Yuu Kawakami, Taro Yajima, Shuichi Miyatake, and
    Anna Nguno. 2013. Mineralogical mapping of southern Namibia by application
    of continuum-removal MSAM method to the HyMap data. Int. J. Remote Sens.
    34, 15 (August 2013), 5282-5295.
    '''
    # The modifications to the `spectral_angles` function were contributed by
    # Christian Mielke.

    assert members.shape[1] == data.shape[2], \
        'Matrix dimensions are not aligned.'

    (M, N, B) = data.shape
    m = np.array(members, np.float64)
    C = m.shape[0]

    # Normalize endmembers
    for i in range(C):
        # Fisher z trafo type operation
        m[i] -= np.mean(m[i])
        m[i] /= np.sqrt(m[i].dot(m[i]))

    angles = np.zeros((M, N, C), np.float64)

    for i in range(M):
        for j in range(N):
            #Fisher z trafo type operation
            v = data[i, j] - np.mean(data[i, j])
            v /= np.sqrt(v.dot(v))
            v = np.clip(v, -1, 1)
            for k in range(C):
                # Calculate Mineral Index according to Oshigami et al.
                # (Intnl. J. of Remote Sens. 2013)
                a = np.clip(v.dot(m[k]), -1, 1)
                angles[i,j,k]= 1.0 - np.arccos(a) / (math.pi / 2)
    return angles

def noise_from_diffs(X, direction='lowerright'):
    '''Estimates noise statistcs by taking differences of adjacent pixels.

    Arguments:

        `X` (np.ndarray):

            The data from which to estimage noise statistics. `X` should have
            shape `(nrows, ncols, nbands`).

        `direction` (str, default "lowerright"):

            The pixel direction along which to calculate pixel differences.
            Must be one of the following:

                'lowerright':
                    Take difference with pixel diagonally to lower right
                'lowerleft':
                     Take difference with pixel diagonally to lower right
                'right':
                    Take difference with pixel to the right
                'lower':
                    Take differenece with pixel below

    Returns a :class:`~spectral.algorithms.algorithms.GaussianStats` object.
    '''
    if direction.lower() not in ['lowerright', 'lowerleft', 'right', 'lower']:
        raise ValueError('Invalid `direction` value.')
    if direction == 'lowerright':
        deltas = X[:-1, :-1, :] - X[1:, 1:, :]
    elif direction == 'lowerleft':
        deltas = X[:-1, 1:, :] - X[1:, :-1, :]
    elif direction == 'right':
        deltas = X[:, :-1, :] - X[:, 1:, :]
    else:
        deltas = X[:-1, :, :] - X[1:, :, :]

    stats = calc_stats(deltas)
    stats.cov /= 2.0
    return stats

class MNFResult(object):
    '''Result object returned by :func:`~spectral.algorithms.algorithms.mnf`.

    This object contains data associates with a Minimum Noise Fraction
    calculation, including signal and noise statistics, as well as the
    Noise-Adjusted Principal Components (NAPC). This object can be used to
    denoise image data or to reduce its dimensionality.
    '''
    def __init__(self, signal, noise, napc):
        '''
        Arguments:

            `signal` (:class:`~spectral.GaussianStats`):

                Signal statistics

            `noise` (:class:`~spectral.GaussianStats`):

                Noise statistics

            `napc` (:class:`~spectral.PrincipalComponents`):

                Noise-Adjusted Pricipal Components
        '''
        self.signal = signal
        self.noise = noise
        self.napc = napc

    def _num_from_kwargs(self, **kwargs):
        '''Returns number of components to retain for the given kwargs.'''
        for key in kwargs:
            if key not in ('num', 'snr'):
                raise Exception('Keyword not recognized.')
        num = kwargs.get('num', None)
        snr = kwargs.get('snr', None)
        if num == snr == None:
            raise Exception('Must specify either `num` or `snr` keyword.')
        if None not in (num, snr):
            raise Exception('Can not specify both `num` and `snr` keywords.')
        if snr is not None:
            num = self.num_with_snr(snr)
        return num

    def denoise(self, X, **kwargs):
        '''Returns a de-noised version of `X`.

        Arguments:

            `X` (np.ndarray):

                Data to be de-noised. Can be a single pixel or an image.

        One (and only one) of the following keywords must be specified:

            `num` (int):

                Number of Noise-Adjusted Principal Components to retain.

            `snr` (float):

                Threshold signal-to-noise ratio (SNR) to retain.

        Returns denoised image data with same shape as `X`.

        Note that calling this method is equivalent to calling the
        `get_denoising_transform` method with same keyword and applying the
        returned transform to `X`. If you only intend to denoise data with the
        same parameters multiple times, then it is more efficient to get the
        denoising transform and reuse it, rather than calling this method
        multilple times.
        '''
        f = self.get_denoising_transform(**kwargs)
        return f(X)

    def get_denoising_transform(self, **kwargs):
        '''Returns a function for denoising image data.

        One (and only one) of the following keywords must be specified:

            `num` (int):

                Number of Noise-Adjusted Principal Components to retain.

            `snr` (float):

                Threshold signal-to-noise ratio (SNR) to retain.

        Returns a callable :class:`~spectral.algorithms.transforms.LinearTransform`
        object for denoising image data.
        '''
        N = self._num_from_kwargs(**kwargs)
        V = self.napc.eigenvectors
        Vr = np.array(V)
        Vr[:, N:] = 0.
        f = LinearTransform(self.noise.sqrt_cov.dot(Vr).dot(V.T) \
			    .dot(self.noise.sqrt_inv_cov),
                            pre=-self.signal.mean,
                            post=self.signal.mean)
        return f

    def reduce(self, X, **kwargs):
        '''Reduces dimensionality of image data.

        Arguments:

            `X` (np.ndarray):

                Data to be reduced. Can be a single pixel or an image.

        One (and only one) of the following keywords must be specified:

            `num` (int):

                Number of Noise-Adjusted Principal Components to retain.

            `snr` (float):

                Threshold signal-to-noise ratio (SNR) to retain.

        Returns a verions of `X` with reduced dimensionality.

        Note that calling this method is equivalent to calling the
        `get_reduction_transform` method with same keyword and applying the
        returned transform to `X`. If you intend to denoise data with the
        same parameters multiple times, then it is more efficient to get the
        reduction transform and reuse it, rather than calling this method
        multilple times.
        '''
        f = self.get_reduction_transform(**kwargs)
        return f(X)

    def get_reduction_transform(self, **kwargs):
        '''Reduces dimensionality of image data.

        One (and only one) of the following keywords must be specified:

            `num` (int):

                Number of Noise-Adjusted Principal Components to retain.

            `snr` (float):

                Threshold signal-to-noise ratio (SNR) to retain.

        Returns a callable :class:`~spectral.algorithms.transforms.LinearTransform`
        object for reducing the dimensionality of image data.
        '''
        N = self._num_from_kwargs(**kwargs)
        V = self.napc.eigenvectors
        f = LinearTransform(V[:, :N].T.dot(self.noise.sqrt_inv_cov),
                            pre=-self.signal.mean)
        return f

    def num_with_snr(self, snr):
        '''Returns the number of components with SNR >= `snr`.'''
        return np.sum(self.napc.eigenvalues >= (snr + 1))

def mnf(signal, noise):
    '''Computes Minimum Noise Fraction / Noise-Adjusted Principal Components.

    Arguments:

        `signal` (:class:`~spectral.algorithms.algorithms.GaussianStats`):

            Estimated signal statistics

        `noise` (:class:`~spectral.algorithms.algorithms.GaussianStats`):

            Estimated noise statistics

    Returns an :class:`~spectral.algorithms.algorithms.MNFResult` object,
    containing the Noise-Adjusted Principal Components (NAPC) and methods for
    denoising or reducing dimensionality of associated data.

    The Minimum Noise Fraction (MNF) is similar to the Principal Components
    transformation with the difference that the Principal Components associated
    with the MNF are ordered by descending signal-to-noise ratio (SNR) rather
    than overall image variance. Note that the eigenvalues of the NAPC are
    equal to one plus the SNR in the transformed space (since noise has
    whitened unit variance in the NAPC coordinate space).

    Example:

        >>> data = open_image('92AV3C.lan').load()
        >>> signal = calc_stats(data)
        >>> noise = noise_from_diffs(data[117: 137, 85: 122, :])
        >>> mnfr = mnf(signal, noise)

        >>> # De-noise the data by eliminating NAPC components where SNR < 10.
        >>> # The de-noised data will be in the original coordinate space (at
        >>> # full dimensionality).
        >>> denoised = mnfr.denoise(data, snr=10)

        >>> # Reduce dimensionality, retaining NAPC components where SNR >= 10.
        >>> reduced = mnfr.reduce(data, snr=10)

        >>> # Reduce dimensionality, retaining top 50 NAPC components.
        >>> reduced = mnfr.reduce(data, num=50)

    References:

        Lee, James B., A. Stephen Woodyatt, and Mark Berman. "Enhancement of
        high spectral resolution remote-sensing data by a noise-adjusted
        principal components transform." Geoscience and Remote Sensing, IEEE
        Transactions on 28.3 (1990): 295-304.
    '''
    C = noise.sqrt_inv_cov.dot(signal.cov).dot(noise.sqrt_inv_cov)
    (L, V) = np.linalg.eig(C)
    # numpy says eigenvalues may not be sorted so we'll sort them, if needed.
    if not np.alltrue(np.diff(L) <= 0):
        ii = list(reversed(np.argsort(L)))
        L = L[ii]
        V = V[:, ii]
    wstats = GaussianStats(mean=np.zeros_like(L), cov=C)
    napc = PrincipalComponents(L, V, wstats)
    return MNFResult(signal, noise, napc)

def ppi(X, niters, threshold=0, centered=False, start=None, display=0,
        **imshow_kwargs):
    '''Returns pixel purity indices for an image.

    Arguments:

        `X` (ndarray):

            Image data for which to calculate pixel purity indices

        `niters` (int):

            Number of iterations to perform. Each iteration corresponds to a
            projection of the image data onto a random unit vector.

        `threshold` (numeric):

            If this value is zero, only the two most extreme pixels will have
            their indices incremented for each random vector. If the value is
            greater than zero, then all pixels whose projections onto the
            random vector are with `threshold` data units of either of the two
            extreme pixels will also have their indices incremented.

        `centered` (bool):

            If True, then the pixels in X are assumed to have their mean
            already subtracted; otherwise, the mean of `X` will be computed
            and subtracted prior to computing the purity indices.

        `start` (ndarray):

            An optional array of initial purity indices. This can be used to
            continue computing PPI values after a previous call to `ppi` (i.e.,
            set `start` equal to the return value from a previou call to `ppi`.
            This should be an integer-valued array whose dimensions are equal
            to the first two dimensions of `X`.

        `display` (integer):

            If set to a postive integer, a :class:`~spectral.graphics.spypylab.ImageView`
            window will be opened and dynamically display PPI values as the
            function iterates. The value specifies the number of PPI iterations
            between display updates. It is recommended to use a value around
            100 or higher. If the `stretch` keyword (see :func:`~spectral.graphics.graphics.get_rgb`
            for meaning) is not provided, a default stretch of (0.99, 0.999)
            is used.

    Return value:

        An ndarray of integers that represent the pixel purity indices of the
        input image. The return array will have dimensions equal to the first
        two dimensions of the input image.

    Keyword Arguments:

        Any keyword accepted by :func:`~spectral.graphics.spypylab.imshow`.
        These keywords will be passed to the image display and only have an
        effect if the `display` argument is nonzero.

    This function can be interruped with a KeyboardInterrupt (ctrl-C), in which
    case, the most recent value of the PPI array will be returned. This can be
    used in conjunction with the `display` argument to view the progression of
    the PPI values until they appear stable, then terminate iteration using
    ctrl-C.

    References:

    Boardman J.W., Kruse F.A, and Green R.O., "Mapping Target Signatures via
    Partial Unmixing of AVIRIS Data," Pasadena, California, USA, 23 Jan 1995,
    URI: http://hdl.handle.net/2014/33635
    '''
    if display is not None:
        if not isinstance(display, Integral) or isinstance(display, bool) or \
            display < 0:
            msg = '`display` argument must be a non-negative integer.'
            raise ValueError(msg)

    if not centered:
        stats = calc_stats(X)
        X = X - stats.mean

    shape = X.shape
    X = X.reshape(-1, X.shape[-1])
    nbands = X.shape[-1]

    fig = None
    updating = False

    if start is not None:
        counts = np.array(start.ravel())
    else:
        counts = np.zeros(X.shape[0], dtype=np.uint32)

    if 'stretch' not in imshow_kwargs:
        imshow_kwargs['stretch'] = (0.99, 0.999)

    msg = 'Running {0} pixel purity iterations...'.format(niters)
    spy._status.display_percentage(msg)

    try:
        for i in range(niters):
            r = np.random.rand(nbands) - 0.5
            r /= np.sqrt(np.sum(r * r))
            s = X.dot(r)
            imin = np.argmin(s)
            imax = np.argmax(s)

            updating = True
            if threshold == 0:
                # Only the two extreme pixels are incremented
                counts[imin] += 1
                counts[imax] += 1
            else:
                # All pixels within threshold distance from the two extremes
                counts[s >= (s[imax] - threshold)] += 1
                counts[s <= (s[imin] + threshold)] += 1
            updating = False

            if display > 0 and (i + 1) % display == 0:
                if fig is not None:
                    fig.set_data(counts.reshape(shape[:2]), **imshow_kwargs)
                else:
                    fig = spy.imshow(counts.reshape(shape[:2]), **imshow_kwargs)
                fig.set_title('PPI ({} iterations)'.format(i + 1))

            if not (i + 1) % 10:
                spy._status.update_percentage(100 * (i + 1) / niters)

    except KeyboardInterrupt:
        spy._status.end_percentage('interrupted')
        if not updating:
            msg = 'KeyboardInterrupt received. Returning pixel purity ' \
              'values after {0} iterations.'.format(i)
            spy._status.write(msg)
            return counts.reshape(shape[:2])
        else:
            msg = 'KeyboardInterrupt received during array update. PPI ' \
              'values may be corrupt. Returning None'
            spy._status.write(msg)
            return None

    spy._status.end_percentage()

    return counts.reshape(shape[:2])


def smacc(spectra, min_endmembers=None, max_residual_norm=float('Inf')):
    '''Returns SMACC decomposition (H = F * S + R) matrices for an image or
    array of spectra.

    Let `H` be matrix of shape NxB, where B is number of bands, and N number of
    spectra, then if `spectra` is of the same shape, `H` will be equal to `spectra`.
    Otherwise, `spectra` is assumed to be 3D spectral image, and it is reshaped
    to match shape of `H`.

    Arguments:

        `spectra` (ndarray):

            Image data for which to calculate SMACC decomposition matrices.

        `min_endmembers` (int):

            Minimal number of endmembers to find. Defaults to rank of `H`,
            computed numerically with `numpy.linalg.matrix_rank`.

        `max_residual_norm`:

            Maximum value of residual vectors' norms. Algorithm will keep finding
            new endmembers until max value of residual norms is less than this
            argument. Defaults to float('Inf')

    Returns:
        3 matrices, S, F and R, such that H = F * S + R (but it might not always hold).
        F is matrix of expansion coefficients of shape N x num_endmembers.
        All values of F are equal to, or greater than zero.
        S is matrix of endmember spectra, extreme vectors, of shape num_endmembers x B.
        R is matrix of residuals of same shape as H (N x B).

        If values of H are large (few tousands), H = F * S + R, might not hold,
        because of numeric errors. It is advisable to scale numbers, by dividing
        by 10000, for example. Depending on how accurate you want it to be,
        you can check if H is really strictly equal to F * S + R,
        and adjust R: R = H - np.matmul(F, S).

    References:

        John H. Gruninger, Anthony J. Ratkowski, and Michael L. Hoke "The sequential
        maximum angle convex cone (SMACC) endmember model", Proc. SPIE 5425, Algorithms
        and Technologies for Multispectral, Hyperspectral, and Ultraspectral Imagery X,
        (12 August 2004); https://doi.org/10.1117/12.543794
    '''
    # Indices of vectors in S.
    q = []

    H = spectra if len(spectra.shape) == 2 else spectra.reshape(
        (spectra.shape[0] * spectra.shape[1], spectra.shape[2]))
    R = H
    Fs = []

    F = None
    S = None

    if min_endmembers is None:
        min_endmembers = np.linalg.matrix_rank(H)

    # Add the longest vector to q.
    residual_norms = np.sqrt(np.einsum('ij,ij->i', H, H))
    current_max_residual_norm = np.max(residual_norms)

    if max_residual_norm is None:
        max_residual_norm = current_max_residual_norm / min_endmembers

    while len(q) < min_endmembers or current_max_residual_norm > max_residual_norm:
        q.append(np.argmax(residual_norms))
        n = len(q) - 1
        # Current basis vector.
        w = R[q[n]]
        # Temporary to be used for projection calculation.
        wt = w / (np.dot(w, w))
        # Calculate projection coefficients.
        On = np.dot(R, wt)
        alpha = np.ones(On.shape, dtype=np.float64)
        # Make corrections to satisfy convex cone conditions.
        # First correct alphas for oblique projection when needed.
        for k in range(len(Fs)):
            t = On * Fs[k][q[n]]
            # This is not so important for the algorithm itself.
            # These values correpond to values where On == 0.0, and these
            # will be zeroed out below. But to avoid divide-by-zero warning
            # we set small values instead of zero.
            t[t == 0.0] = 1e-10
            np.minimum(Fs[k]/t, alpha, out=alpha)
        # Clip negative projection coefficients.
        alpha[On <= 0.0] = 0.0
        # Current extreme vector should always be removed completely.
        alpha[q[n]] = 1.0
        # Calculate oblique projection coefficients.
        Fn = alpha * On
        # Correction for numerical stability.
        Fn[Fn <= 0.0] = 0.0
        # Remove projection to current basis from R.
        R = R - np.outer(Fn, w)
        # Update projection coefficients.
        for k in range(len(Fs)):
            Fs[k] -= Fs[k][q[n]] * Fn
            # Correction because of numerical problems.
            Fs[k][Fs[k] <= 0.0] = 0.0

        # Add new Fn.
        Fs.append(Fn)

        residual_norms[:] = np.sqrt(np.einsum('ij,ij->i', R, R))
        current_max_residual_norm = np.max(residual_norms)
        print('Found {0} endmembers, current max residual norm is {1:.4f}\r'
            .format(len(q), current_max_residual_norm), end='')

    # Correction as suggested in the SMACC paper.
    for k, s in enumerate(q):
        Fs[k][q] = 0.0
        Fs[k][s] = 1.0

    F = np.array(Fs).T
    S = H[q]

    # H = F * S + R
    return S, F, R



'''
Continuum and continuum removal.

Continuum is defined as convex hull of spectrum.
Continuum is removed from spectra by dividing spectra by its continuum.
That results in values between 0 and 1, where absorption bands are expressed as
drops below 1. It is usefull for comparing and classification based on
absorption bands and indifferent to scale.

References:
    Clark, R.N. and Roush, L. (1984) Reflectance Spectroscopy Quantitative Analysis
    Techniques for Remote Sensing Applications. Journal of Geophysical Research,
    89, 6329-6340. http://dx.doi.org/10.1029/JB089iB07p06329

    Jiwei Bai, et al., "Classification methods of the hyperspectralimage based
    on the continuum-removed," Proc. SPIE 4897, Multispectral and Hyperspectral
    Remote Sensing Instruments and Applications, (16 June 2003);
    doi: 10.1117/12.466729

    Lehnert, Lukas & Meyer, Hanna & Obermeier, Wolfgang & Silva, Brenner & Regeling,
    Bianca & Thies, Boris & Bendix, Jorg. (2019). Hyperspectral Data Analysis in R:
    The hsdar Package. Journal of statistical software. 89. 1-23. 10.18637/jss.v089.i12.
'''

from __future__ import absolute_import, division, print_function, unicode_literals

import logging
import numpy as np

import spectral as spy
from ..utilities.errors import has_nan, NaNValueError


def _segment_concave_region(spectrum, bands, indices, ind_fill, ibegin, iend):
    # Here we don't search for local maxima w.r.t. line that connects ends of this region.
    # That is behavior of the hsdar. It also makes more sense in the context of
    # hyperspectral image analysis. We are already not convex, and we can't
    # include all points that make result quasi-convex, since there will be too
    # many of them, often right one after another. However, filtering local maxima,
    # below, will make result quasi-convex.
    # Notice that we are using >=, not strict >. That will include maxima that
    # are flat, that stretch few points. It will also include local minima,
    # but just as with local maxima that are too low below slope line, these
    # will be filtered out.
    is_maximum = np.logical_and(spectrum[ibegin+1:iend-1] >= spectrum[ibegin:iend-2],
                                spectrum[ibegin+1:iend-1] >= spectrum[ibegin+2:iend])

    # Get local maxima indices. (Note that where return tuple for each dimension).
    lmi = np.where(is_maximum)[0]

    # No local maxima, return.
    if len(lmi) == 0:
        return ind_fill

    # Make it relative to input array - spectrum.
    lmi += ibegin + 1

    # Get local maxima.
    local_maxima = spectrum[lmi]

    # Filter those maxima that cause slope between them to change direction.
    # This makes remaining maxima, satisfy quasy-convexity condition.
    slope_dir = spectrum[iend-1] - spectrum[ibegin]
    filtered_indices = []
    if slope_dir >= 0.0:
        last_included_value = spectrum[ibegin]
        for i in range(len(local_maxima)):
            lm = local_maxima[i]
            if lm > last_included_value:
                filtered_indices.append(lmi[i])
                last_included_value = lm
    else:
        # Slope is negative. Start from back.
        last_included_value = spectrum[iend-1]
        for i in range(len(local_maxima) - 1, -1, -1):
            lm = local_maxima[i]
            if lm > last_included_value:
                filtered_indices.append(lmi[i])
                last_included_value = lm
        filtered_indices.reverse()

    # Take only valid local maxima indices.
    lmi = filtered_indices

    # If there is no valid local maxima indices, return.
    if len(lmi) == 0:
        return ind_fill

    # Add indices to result, and process subregions between them with convex hull
    # algorithm, to make sure all input points and below resulting hull.
    next_ibegin = ibegin
    for i in lmi:
        # There should be at least 1 point between edges, to call _find_indices_in_range.
        # However, these are to local maxima, and if there is one point between them,
        # it must be below both. So only for two points inside region borders
        # call _find_indices_in_range.
        if i > next_ibegin + 2:
            # Put hull around points in subregion.
            ind_fill = _find_indices_in_range(
                spectrum, bands, False, indices, ind_fill, next_ibegin, i + 1)
        indices[ind_fill] = i
        ind_fill += 1
        next_ibegin = i

    # Don't miss the last range.
    ind_fill = _find_indices_in_range(
        spectrum, bands, False, indices, ind_fill, lmi[-1], iend)

    return ind_fill


def _find_indices_in_range(spectrum, bands, segmented, indices, ind_fill, ibegin, iend):
    iendi = iend - 1

    # We search for maximum, but not from the x axis.
    # We search for maximum w.r.t to axis represented by line connecting
    # first and last point (of this iteration).

    # First find normal to new axis. Swap x and y, and negate new x.
    # If we negate x instead of y, normal will always point upward.
    naxis_y = bands[iendi] - bands[ibegin]
    naxis_x = spectrum[ibegin] - spectrum[iendi]

    # Don't literally compute distance from the axis. Rather, calculate dot products
    # of points with the normal, and find the largest. The largest dot product (it does not have to be positive)
    # is the one that goes more in the direction of normal than others. To get the distance,
    # we could divide each dot product by norm/length of the normal. But that is constant,
    # and does not effect which one is maximum.
    # Note that here we include first point of the range, but not last.
    imax = np.argmax(bands[ibegin:iendi] * naxis_x +
                     spectrum[ibegin:iendi] * naxis_y) + ibegin

    # If first point is maximum, then all others are "below" the axis,
    # which means this is concave region.
    if imax == ibegin:
        # If we are in segmented upper hull mode, then segment concave region.
        # For that to make sense, we need at least 3 elements between edges.
        if segmented and iend - ibegin > 5:
            ind_fill = _segment_concave_region(
                spectrum, bands, indices, ind_fill, ibegin, iend)
        return ind_fill

    # Repeat same procedure on the left side, if there are enough points left.
    # At least 1 is required between first and last point in range.
    if imax > ibegin + 1:
        ind_fill = _find_indices_in_range(
            spectrum, bands, segmented, indices, ind_fill, ibegin, imax + 1)

    # Push middle index.
    indices[ind_fill] = imax
    ind_fill += 1

    # Repeat for the right side.
    if imax < iend - 2:
        ind_fill = _find_indices_in_range(
            spectrum, bands, segmented, indices, ind_fill, imax, iend)

    return ind_fill


def _find_continuum_points_recursive(spectrum, bands, segmented, indices):
    n = len(spectrum)
    indices[0] = 0
    ind_fill = 1

    ind_fill = _find_indices_in_range(
        spectrum, bands, segmented, indices, ind_fill, 0, n)
    indices[ind_fill] = n - 1
    indices = indices[:ind_fill + 1]

    return (bands[indices], spectrum[indices])


def _process_continuum(spectra, bands, remove_continuum, segmented, out):
    if not isinstance(spectra, np.ndarray):
        raise TypeError('Expected spectra to be a numpy.ndarray.')
    if not isinstance(bands, np.ndarray):
        raise TypeError('Expected bands to be a numpy.ndarray.')
    if out is not None and not isinstance(out, np.ndarray):
        raise TypeError('Expected out to be a numpy.ndarray or None.')
    if len(spectra.shape) not in (1, 2, 3):
        raise ValueError('Expected spectra to be 1d, 2d, or 3d array.')
    if len(bands.shape) != 1:
        raise ValueError('Expected bands to be 1d array.')
    if out is not None and not np.array_equal(out.shape, spectra.shape):
        raise ValueError('Expected out to be same shape as spectra.')

    out = np.empty_like(spectra) if out is None else out

    # In case we remove continuum, always divide out by continuum,
    # to avoid creating additional temporary array.
    if spectra is not out and remove_continuum:
        out[:] = spectra[:]

    original_shape = spectra.shape
    nbands = original_shape[-1]

    interp = np.interp
    indices = np.empty(nbands, np.int64)

    if len(spectra.shape) == 1:
        points = _find_continuum_points_recursive(
            spectra, bands, segmented, indices)
        continuum = interp(bands, points[0], points[1])
        if remove_continuum:
            out /= continuum
        else:
            out[:] = continuum
    elif len(spectra.shape) == 2:
        for i in range(spectra.shape[0]):
            points = _find_continuum_points_recursive(
                spectra[i], bands, segmented, indices)
            continuum = interp(bands, points[0], points[1])
            if remove_continuum:
                out[i, :] /= continuum
            else:
                out[i, :] = continuum
    else:
        for i in range(spectra.shape[0]):
            for j in range(spectra.shape[1]):
                points = _find_continuum_points_recursive(
                    spectra[i, j], bands, segmented, indices)
                continuum = interp(bands, points[0], points[1])
                if remove_continuum:
                    out[i, j, :] /= continuum
                else:
                    out[i, j, :] = continuum

    return out


def continuum_points(spectrum, bands, mode='convex'):
    '''Returns points of spectra that belong to it's continuum.

    Arguments:

        `spectrum` (:class:`numpy.ndarray`)

            1d :class:`numpy.ndarray` holding spectral signature.

        `bands` (:class:`numpy.ndarray`):

            1d :class:`numpy.ndarray`, holding band values of spectra.
            Length of `bands` should be the same as `spectrum`.
            Note that bands should be sorted in ascending order (which is often
            not the case with AVIRIS), otherwise unexpected results could occure.

        `mode` (string, default 'convex'):

            Default mode is 'convex' which returns convex upper hull of the
            spectrum. Another supported mode is 'segmented' which builds
            segmented upper hull. This is usefull to identify more detailed
            contour of the spectrum, but without strong absorption bands.

    Returns:

        2-tuple, with each element being :class:`numpy.ndarray`.
        First element contains reflectance values of points that belong to
        continuum. Second element contains corresponding bands.
        By applying linear interpolation to this data as x and y, we get
        continuum of spectrum. However this function is particularly useful to
        applying other interpolations or any other processing on these points.
    '''
    if not isinstance(spectrum, np.ndarray):
        raise TypeError('Expected spectra to be a numpy.ndarray.')
    if not isinstance(bands, np.ndarray):
        raise TypeError('Expected bands to be a numpy.ndarray.')
    if len(spectrum.shape) != 1:
        raise ValueError('Expected spectra to be 1d array.')
    if len(bands.shape) != 1:
        raise ValueError('Expected bands to be 1d array.')

    indices = np.empty_like(spectrum, dtype='int64')
    return _find_continuum_points_recursive(spectrum, bands, mode == 'segmented', indices)


def spectral_continuum(spectra, bands, mode='convex', out=None):
    '''Returns continua of spectra.
    Continuum is defined as convex hull of spectra.

    Arguments:

        `spectra` (:class:`numpy.ndarray`)

            Can be 1d, 2d or 3d :class:`numpy.ndarray`, where last dimension
            holds individual spectra.

        `bands` (:class:`numpy.ndarray`):

            1d :class:`numpy.ndarray`, holding band values of spectra.
            Length of `bands` should be the same as last dimension of `spectra`.
            Note that bands should be sorted in ascending order (which is often
            not the case with AVIRIS), otherwise unexpected results could occure.

        `mode` (string, default 'convex'):

            Default mode is 'convex' which returns convex upper hull of the
            spectrum. Another supported mode is 'segmented' which builds
            segmented upper hull. This is usefull to identify more detailed
            contour of the spectrum, but without strong absorption bands.

        `out` (:class:`numpy.ndarray`, default None):

            If provided, it must have same type and same shape as `spectra`,
            and it will hold the result, and will be returned as result of this
            function.

    Returns:

        A :class:`numpy.ndarray` of continua for each spectrum in spectra.
        It same type and shape as spectra. If `out` is provided, `out` will be
        returned.
    '''
    return _process_continuum(spectra, bands, False, mode == 'segmented', out)


def remove_continuum(spectra, bands, mode='convex', out=None):
    '''Returns spectra with continuum removed.
    Continuum is defined as convex hull of spectra. Continuum is removed from
    spectra by deviding spectra by its continuum.

    Arguments:

        `spectra` (:class:`numpy.ndarray`)

            Can be 1d, 2d or 3d :class:`numpy.ndarray`, where last dimension
            holds individual spectra.

        `bands` (:class:`numpy.ndarray`):

            1d :class:`numpy.ndarray`, holding band values of spectra.
            Length of `bands` should be the same as last dimension of `spectra`.
            Note that bands should be sorted in ascending order (which is often
            not the case with AVIRIS), otherwise unexpected results could occure.

        `mode` (string, default 'convex'):

            Default mode is 'convex' which removes convex upper hull of the
            spectrum. Another supported mode is 'segmented' which removes
            segmented upper hull. This is usefull to identify two or more small
            features instead of one large feature.

        `out` (:class:`numpy.ndarray`, default None):

            If provided, it must have type `np.float64` and same shape as
            `spectra`, and it will hold the result, and will be returned as
            result of this function.

    Returns:

        A :class:`numpy.ndarray` of continua for in spectrum in spectra.
        It type `np.float64` and same shape as spectra. If `out` is provided,
        `out` will be returned.
    '''
    if out is not None and out.dtype != np.float64:
        raise ValueError('Expected out to have dtype float64. '
                         'Results of continuum removal are floating point numbers.')
    return _process_continuum(spectra, bands, True, mode == 'segmented', out)



'''
Code for reading and managing ECOSTRESS spectral library data.
'''

from __future__ import absolute_import, division, print_function, unicode_literals

import itertools
import logging

from spectral.utilities.python23 import IS_PYTHON3
from .aster import AsterDatabase, Signature

if IS_PYTHON3:
    readline = lambda fin: fin.readline()
    open_file = lambda filename: open(filename, encoding='iso-8859-1')
else:
    readline = lambda fin: fin.readline().decode('iso-8859-1')
    open_file = lambda filename: open(filename)


def read_ecostress_file(filename):
    '''Reads an ECOSTRESS v1 spectrum file.'''

    logger = logging.getLogger('spectral')
    lines = open_file(filename).readlines()
    if not IS_PYTHON3:
        lines = [line.decode('iso-8859-1') for line in lines]

    metaline_to_pair = lambda line: [x.strip() for x in line.split(':', 1)]

    s = Signature()

    # Read sample metadata
    for i in itertools.count():
        if lines[i].strip().startswith('Measurement'):
            break
        pair = metaline_to_pair(lines[i])
        try:
            s.sample[pair[0].lower()] = pair[1]
        except:
            logger.error('Failed to parse line: {}: {}'.format(i, lines[i]))
            raise

    # Read measurment metadata
    for j in itertools.count(i):
        if len(lines[j].strip()) == 0:
            break
        pair = metaline_to_pair(lines[j])
        s.measurement[pair[0].lower()] = pair[1]

    # Read signature spectrum
    pairs = []
    for line in lines[j:]:
        line = line.strip()
        if len(line) == 0:
            continue
        pair = line.split()
        nItems = len(pair)

        # Try to handle invalid values on signature lines
        if nItems == 1:
            logger.info('Skipping single item (%s) on signature line for %s',
                  pair[0], filename)
            continue
        elif nItems > 2:
            logger.info('Skipping more than 2 values on signature line for %s',
                        filename)
            continue
        try:
            x = float(pair[0])
        except:
            logger.info('Corrupt signature line in file %s', filename)
        if x == 0:
            logger.info('Skipping zero wavelength value in file %s', filename)
            continue
        elif x < 0:
            logger.info('Skipping negative wavelength value in file %s',
                        filename)
            continue

        pairs.append(pair)

    [x, y] = [list(v) for v in zip(*pairs)]

    # Make sure wavelengths are ascending
    if float(x[0]) > float(x[-1]):
        x.reverse()
        y.reverse()
    s.x = [float(val) for val in x]
    s.y = [float(val) for val in y]
    s.measurement['first x value'] = x[0]
    s.measurement['last x value'] = x[-1]
    s.measurement['number of x values'] = len(x)

    return s

class EcostressDatabase(AsterDatabase):
    '''A relational database to manage ECOSTRESS spectral library data.'''

    @classmethod
    def create(cls, filename, data_dir=None):
        '''Creates an ECOSTRESS relational database by parsing ECOSTRESS data files.

        Arguments:

            `filename` (str):

                Name of the new sqlite database file to create.

            `data_dir` (str):

                Path to the directory containing ECOSTRESS library data files. If
                this argument is not provided, no data will be imported.

        Returns:

            An :class:`~spectral.database.EcostressDatabase` object.

        Example::

            >>> EcostressDatabase.create("ecostress.db", "./eco_data_ver1/")

        This is a class method (it does not require instantiating an
        EcostressDatabase object) that creates a new database by parsing all of the
        files in the ECOSTRESS library data directory.  Normally, this should only
        need to be called once.  Subsequently, a corresponding database object
        can be created by instantiating a new EcostressDatabase object with the
        path the database file as its argument.  For example::

            >>> from spectral.database.ecostress import EcostressDatabase
            >>> db = EcostressDatabase("~/ecostress.db")
        '''
        import os
        if os.path.isfile(filename):
            raise Exception('Error: Specified file already exists.')
        db = cls()
        db._connect(filename)
        for schema in cls.schemas:
            db.cursor.execute(schema)
        if data_dir:
            db._import_files(data_dir)
        return db

    def read_file(self, filename):
        return read_ecostress_file(filename)

    def _import_files(self, data_dir, ignore=None):
        '''Import each file from the ECOSTRESS library into the database.'''
        from glob import glob
        import numpy
        import os

        logger = logging.getLogger('spectral')
        if not os.path.isdir(data_dir):
            raise Exception('Error: Invalid directory name specified.')
        if ignore is not None:
            filesToIgnore = [data_dir + '/' + f for f in ignore]
        else:
            filesToIgnore = []

        numFiles = 0
        numIgnored = 0

        sigID = 1

        class Sig:
            pass
        sigs = []

        for f in glob(data_dir + '/*spectrum.txt'):
            if f in filesToIgnore:
                numIgnored += 1
                continue
            logger.info('Importing ECOSTRESS file %s.', f)
            numFiles += 1
            sig = self.read_file(f)
            s = sig.sample
            if 'particle size' in s:
                if s['particle size'].lower == 'liquid':
                    phase = 'liquid'
                else:
                    phase = 'solid'
            else:
                phase = 'unknown'
                s['particle size'] = 'none'
            if 'sample no.' in s:
                sampleNum = s['sample no.']
            else:
                sampleNum = ''
            subclass = s.get('subclass', 'none')
            if subclass is 'none' and 'genus' in s:
                subclass = s['genus']
            id = self._add_sample(s['name'], s['type'], s['class'], subclass,
                                  s['particle size'], sampleNum, s['owner'],
                                  s['origin'], phase, s['description'])

            instrument = os.path.basename(f).split('.')[1]
            environment = 'lab'
            m = sig.measurement

            # Correct numerous mispellings of "reflectance" and "transmittance"
            yUnit = m['y units']
            if yUnit.find('reflectence') > -1:
                yUnit = 'reflectance (percent)'
            elif yUnit.find('trans') == 0:
                yUnit = 'transmittance (percent)'
            measurement = m['measurement']
            if measurement[0] == 't':
                measurement = 'transmittance'
            self._add_signature(id, -1, instrument, environment, measurement,
                                m['x units'], yUnit, m['first x value'],
                                m['last x value'], sig.x, sig.y)
        if numFiles == 0:
            logger.warning('No ECOSTRESS data files were found in "%s".',
                           data_dir)
        else:
            logger.info('Processed %d ECOSTRESS files.', numFiles)
        if numIgnored > 0:
            msg = 'Ignored the following %d bad files:' % (numIgnored)
            for f in filesToIgnore:
                msg += '\n\t' + f
            logger.info(msg)

        return sigs



'''
Runs unit tests for various target detectors.

To run the unit tests, type the following from the system command line:

    # python -m spectral.tests.detectors
'''

from __future__ import absolute_import, division, print_function, unicode_literals

import numpy as np
from numpy.testing import assert_allclose

import spectral as spy
from spectral.tests.spytest import SpyTest


class MatchedFilterTest(SpyTest):
    def setup(self):
        from spectral.algorithms.detectors import MatchedFilter
        self.data = spy.open_image('92AV3C.lan').load()
        self.background = spy.calc_stats(self.data)
        self.target_ij = [33, 87]
#        self.target = self.data[33, 87]
        (i, j) = self.target_ij
        self.mf = MatchedFilter(self.background, self.data[i, j])

    def test_mf_bg_eq_zero(self):
        '''Matched Filter response of background should be zero.'''
        (i, j) = self.target_ij
        np.testing.assert_approx_equal(self.mf(self.background.mean), 0)
        
    def test_mf_target_eq_one(self):
        '''Matched Filter response of target should be one.'''
        from spectral.algorithms.detectors import matched_filter
        (i, j) = self.target_ij
        target = self.data[i, j]
        mf = matched_filter(self.data, target, self.background)
        np.testing.assert_approx_equal(mf[i, j], 1)

    def test_mf_target_no_bg_eq_one(self):
        '''Matched Filter response of target should be one.'''
        from spectral.algorithms.detectors import matched_filter
        (i, j) = self.target_ij
        target = self.data[i, j]
        mf = matched_filter(self.data, target)
        np.testing.assert_approx_equal(mf[i, j], 1)

    def test_mf_target_pixel_eq_one(self):
        '''Matched Filter response of target pixel should be one.'''
        (i, j) = self.target_ij
        np.testing.assert_approx_equal(self.mf(self.data)[i, j], 1)

    def test_mf_windowed_target_eq_one(self):
        '''Windowed Matched Filter response of target pixel should be one.'''
        X = self.data[:10, :10, :]
        ij = (3, 3)
        y = spy.matched_filter(X, X[ij], window=(3,7), cov=self.background.cov)
        np.allclose(1, y[ij])


class RXTest(SpyTest):
    def setup(self):
        self.data = spy.open_image('92AV3C.lan').load()
        self.background = spy.calc_stats(self.data)

    def test_rx_bg_eq_zero(self):
        from spectral.algorithms.detectors import rx, RX
        d = rx(self.data)
        stats = spy.calc_stats(self.data)
        np.testing.assert_approx_equal(rx(stats.mean, background=stats), 0)
        

class ACETest(SpyTest):
    def setup(self):
        self.data = spy.open_image('92AV3C.lan').load()
        self.bg = spy.calc_stats(self.data)
        self.X = self.data[:20, :20, :]

    def test_ace_bg_eq_zero(self):
        '''ACE score of background mean should be zero.'''
        ij = (10, 10)
        y = spy.ace(self.bg.mean, self.X[ij], background=self.bg)
        assert(np.allclose(0, y))
        
    def test_ace_pixel_target_eq_one(self):
        '''ACE score of target should be one for single pixel arg.'''
        ij = (10, 10)
        y = spy.ace(self.X[ij], self.X[ij], background=self.bg)
        assert(np.allclose(1, y))

    def test_ace_novec_pixel_target_eq_one(self):
        '''ACE score of target should be one for single pixel arg.'''
        ij = (10, 10)
        y = spy.ace(self.X[ij], self.X[ij], background=self.bg, vectorize=False)
        assert(np.allclose(1, y))

    def test_ace_target_eq_one(self):
        '''ACE score of target should be one.'''
        ij = (10, 10)
        y = spy.ace(self.X, self.X[ij], background=self.bg)
        assert(np.allclose(1, y[ij]))

    def test_ace_novec_target_eq_one(self):
        '''ACE score (without vectorization) of target should be one.'''
        ij = (10, 10)
        y = spy.ace(self.X, self.X[ij], background=self.bg, vectorize=False)
        assert(np.allclose(1, y[ij]))

    def test_ace_multi_targets_eq_one(self):
        '''ACE score of multiple targets should each be one.'''
        ij1 = (10, 10)
        ij2 = (3, 12)
        y = spy.ace(self.X, [self.X[ij1], self.X[ij2]], background=self.bg)
        assert(np.allclose(1, [y[ij1][0], y[ij2][1]]))

    def test_ace_novec_multi_targets_eq_one(self):
        '''ACE score of multiple targets should each be one.'''
        ij1 = (10, 10)
        ij2 = (3, 12)
        y = spy.ace(self.X, [self.X[ij1], self.X[ij2]], background=self.bg,
                    vectorize=False)
        assert(np.allclose(1, [y[ij1][0], y[ij2][1]]))

    def test_ace_multi_targets_bg_eq_zero(self):
        '''ACE score of background for multiple targets should be one.'''
        ij1 = (10, 10)
        ij2 = (3, 12)
        y = spy.ace(self.bg.mean, [self.X[ij1], self.X[ij2]],
                    background=self.bg)
        assert(np.allclose(0, y))

    def test_ace_subspace_targets_eq_one(self):
        '''ACE score of targets defining target subspace should each be one.'''
        ij1 = (10, 10)
        ij2 = (3, 12)
        y = spy.ace(self.X, np.array([self.X[ij1], self.X[ij2]]),
                    background=self.bg)
        assert(np.allclose(1, [y[ij1], y[ij2]]))

    def test_ace_novec_subspace_targets_eq_one(self):
        '''ACE score of targets defining target subspace should each be one.'''
        ij1 = (10, 10)
        ij2 = (3, 12)
        y = spy.ace(self.X, np.array([self.X[ij1], self.X[ij2]]),
                    background=self.bg, vectorize=False)
        assert(np.allclose(1, [y[ij1], y[ij2]]))

    def test_ace_subspace_bg_eq_zero(self):
        '''ACE score of background for target subspace should be zero.'''
        ij1 = (10, 10)
        ij2 = (3, 12)
        y = spy.ace(self.bg.mean, np.array([self.X[ij1], self.X[ij2]]),
                    background=self.bg)
        assert(np.allclose(0, y))

    def test_ace_windowed_target_eq_one(self):
        '''ACE score of target for windowed background should be one.'''
        ij = (10, 10)
        y = spy.ace(self.X, self.X[ij], window=(3,7), cov=self.bg.cov)
        assert(np.allclose(1, y[ij]))


def run():
    print('\n' + '-' * 72)
    print('Running target detector tests.')
    print('-' * 72)
    for T in [MatchedFilterTest, RXTest, ACETest]:
        T().run()

if __name__ == '__main__':
    from spectral.tests.run import parse_args, reset_stats, print_summary
    parse_args()
    reset_stats()
    run()
    print_summary()





class SpectralDatabase:
    def _connect(self, sqlite_filename):
        '''Establishes a connection to the Specbase sqlite database.'''
        import sqlite3
        self.db = sqlite3.connect(sqlite_filename)
        self.cursor = self.db.cursor()

    def query(self, sql, args=None):
        '''Returns the result of an arbitrary SQL statement.

        Arguments:

            `sql` (str):

                An SQL statement to be passed to the database. Use "?" for
                variables passed into the statement.

            `args` (tuple):

                Optional arguments which will replace the "?" placeholders in
                the `sql` argument.

        Returns:

            An :class:`sqlite3.Cursor` object with the query results.

        Example::

            >>> sql = r'SELECT SpectrumID, Name FROM Samples, Spectra ' +
            ...        'WHERE Spectra.SampleID = Samples.SampleID ' +
            ...        'AND Name LIKE "%grass%" AND MinWavelength < ?'
            >>> args = (0.5,)
            >>> cur = db.query(sql, args)
            >>> for row in cur:
            ...     print row
            ...
            (356, u'dry grass')
            (357, u'grass')
        '''
        if args:
            return self.cursor.execute(sql, args)
        else:
            return self.cursor.execute(sql)

    def print_query(self, sql, args=None):
        '''Prints the text result of an arbitrary SQL statement.

        Arguments:

            `sql` (str):

                An SQL statement to be passed to the database. Use "?" for
                variables passed into the statement.

            `args` (tuple):

                Optional arguments which will replace the "?" placeholders in
                the `sql` argument.

        This function performs the same query function as
        :meth:`spectral.database.SpectralDatabase.query` except query results are
        printed to **stdout** instead of returning a cursor object.

        Example:

            >>> sql = r'SELECT SpectrumID, Name FROM Samples, Spectra ' +
            ...        'WHERE Spectra.SampleID = Samples.SampleID ' +
            ...        'AND Name LIKE "%grass%" AND MinWavelength < ?'
            >>> args = (0.5,)
            >>> db.print_query(sql, args)
            356|dry grass
            357|grass
        '''
        ret = self.query(sql, args)
        for row in ret:
            print("|".join([str(x) for x in row]))




'''
Basic package setup and global imports.
'''

from __future__ import absolute_import, division, print_function, unicode_literals

__version__ = '0.22'

import sys
if sys.byteorder == 'little':
    byte_order = 0   # little endian
else:
    byte_order = 1   # big endian

BSQ = 0
BIL = 1
BIP = 2

from .utilities.errors import SpyException
from .config import SpySettings, spy_colors
settings = SpySettings()

from .spectral import (open_image, load_training_sets, save_training_sets,
                      BandInfo)
from .io import *
from .algorithms import *
from .graphics import *
from .database import *

# Import some submodules into top-level namespace
from .algorithms import detectors

from .spectral import _init
_init()
del _init



'''
Top-level functions & classes.
'''

from __future__ import absolute_import, division, print_function, unicode_literals

import logging
import numbers
import numpy as np
import pickle
import os
from warnings import warn

#from .algorithms.algorithms import TrainingClassSet
#from . import io
#from .io import aviris, envi, erdas, spyfile
#from .io.spyfile import find_file_path, SpyFile

from . import settings

def _init():
    '''Basic configuration of the spectral package.'''
    _setup_logger()
    try:
        global settings
        from .graphics import graphics as spygraphics
        from .graphics import spypylab
        settings.plotter = spypylab
        settings.viewer = spygraphics
    except:
        raise
        warn('Unable to import or configure pylab plotter.  Spectrum plots '
             'will be unavailable.', UserWarning)

    from .utilities import status
    spectral = __import__(__name__.split('.')[0])
    spectral._status = status.StatusDisplay()

def _setup_logger():
    logger = logging.getLogger('spectral')
    logger.setLevel(logging.INFO)
    ch = logging.StreamHandler()
    formatter = logging.Formatter('%(name)s:%(levelname)s: %(message)s')
    ch.setFormatter(formatter)
    logger.addHandler(ch)

class BandInfo:
    '''A BandInfo object characterizes the spectral bands associated with an
    image. All BandInfo member variables are optional.  For *N* bands, all
    members of type <list> will have length *N* and contain float values.

    =================   =====================================   =======
        Member                  Description                     Default
    =================   =====================================   =======
    centers             List of band centers                    None
    bandwidths          List of band FWHM values                None
    centers_stdevs      List of std devs of band centers        None
    bandwidth_stdevs    List of std devs of bands FWHMs         None
    band_quantity       Image data type (e.g., "reflectance")   ""
    band_unit           Band unit (e.g., "nanometer")           ""
    =================   =====================================   =======
    '''
    def __init__(self):
        self.centers = None
        self.bandwidths = None
        self.centers_stdevs = None
        self.bandwidth_stdevs = None
        self.band_quantity = None
        self.band_unit = None


def open_image(file):
    '''
    Locates & opens the specified hyperspectral image.

    Arguments:

        file (str):
            Name of the file to open.

    Returns:

        SpyFile object to access the file.

    Raises:

        IOError.

    This function attempts to determine the associated file type and open the
    file. If the specified file is not found in the current directory, all
    directories listed in the :const:`SPECTRAL_DATA` environment variable will
    be searched until the file is found.  If the file being opened is an ENVI
    file, the `file` argument should be the name of the header file.
    '''
    from . import io
    pathname = io.spyfile.find_file_path(file)

    # Try to open it as an ENVI header file.
    try:
        return io.envi.open(pathname)
    except io.envi.FileNotAnEnviHeader:
        # It isn't an ENVI file so try another file type
        pass
    except:
        raise

    # Maybe it's an Erdas Lan file
    try:
        return io.erdas.open(pathname)
    except:
        pass

    # See if the size is consistent with an Aviris file
    try:
        return io.aviris.open(pathname)
    except:
        pass

    raise IOError('Unable to determine file type or type not supported.')



def save_training_sets(sets, file):
    '''
    Saves a list of TrainingSet objects to a file.  This function assumes
    that all the sets in the list refer to the same image and mask array.
    If that is not the case, this function should not be used.
    '''
    f = open(file, 'w')
    z = array.array([])

    pickle.dump(len(sets), f)
    DumpArray(sets[0].mask, f)
    for s in sets:
        s.mask = z
        s.dump(f)

    f.close()


def load_training_sets(file, image=None):
    '''
    Loads a list of TrainingSet objects from a file.  This function assumes
    that all the sets in the list refer to the same image and mask array.
    If that is not the case, this function should not be used.
    '''
    from .algorithms.algorithms import TrainingClassSet
    ts = TrainingClassSet()
    ts.load(file, image)
    return ts




'''
Runs unit tests of spectral file I/O functions.

The unit tests in this module assume the example file "92AV3C.lan" is in the
spectral data path.  After the file is opened it is saved in various formats
(different combinations of byte order, interleave, and data type) and for each
file written, the new file is opened and known data values are read and checked
to verify they are read properly.

To run the unit tests, type the following from the system command line:

    # python -m spectral.tests.spyfile
'''

from __future__ import division, print_function, unicode_literals

import itertools
import numpy as np
import os


import spectral as spy
from spectral.io.spyfile import find_file_path, FileNotFoundError, SpyFile
from spectral.tests import testdir
from spectral.tests.spytest import SpyTest


def assert_almost_equal(a, b, **kwargs):
    if not np.allclose(a, b, **kwargs):
        raise Exception('NOPE')

class SpyFileTest(SpyTest):
    '''Tests that SpyFile methods read data correctly from files.'''
    def __init__(self, file, datum, value):
        '''
        Arguments:

            `file` (str or `SpyFile`):

                The SpyFile to be tested.  This can be either the name of the
                file or a SpyFile object that has already been opened.

            `datum` (3-tuple of ints):

                (i, j, k) are the row, column and band of the datum to be
                tested. 'i' and 'j' should be at least 10 pixels away from the
                edge of the associated image and `k` should have at least 10
                bands above and below it in the image.

            `value` (int or float):

                The scalar value associated with location (i, j, k) in
                the image.
        '''
        self.file = file
        self.datum = datum
        self.value = value

    def setup(self):
        if isinstance(self.file, SpyFile):
            self.image = self.file
        else:
            self.image = spy.open_image(self.file)

    def test_read_datum(self):
        assert_almost_equal(self.image.read_datum(*self.datum, use_memmap=True),
                            self.value)
        assert_almost_equal(self.image.read_datum(*self.datum, use_memmap=False),
                            self.value)

    def test_read_pixel(self):
        (i, j, k) = self.datum
        assert_almost_equal(self.image.read_pixel(i, j, use_memmap=True)[k],
                            self.value)
        assert_almost_equal(self.image.read_pixel(i, j, use_memmap=False)[k],
                            self.value)

    def test_read_band(self):
        (i, j, k) = self.datum
        assert_almost_equal(self.image.read_band(k, use_memmap=True)[i, j],
                            self.value)
        assert_almost_equal(self.image.read_band(k, use_memmap=False)[i, j],
                            self.value)

    def test_read_bands(self):
        (i, j, k) = self.datum
        bands = (k - 5, k - 2, k, k + 1)
        assert_almost_equal(self.image.read_bands(bands,
                                                  use_memmap=True)[i, j, 2],
                            self.value)
        assert_almost_equal(self.image.read_bands(bands,
                                                  use_memmap=False)[i, j, 2],
                            self.value)

    def test_read_bands_nonascending(self):
        (i, j, k) = self.datum
        bands = (k - 2, k + 1, k, k - 5)
        assert_almost_equal(self.image.read_bands(bands,
                                                  use_memmap=True)[i, j, 2],
                            self.value)
        assert_almost_equal(self.image.read_bands(bands,
                                                  use_memmap=False)[i, j, 2],
                            self.value)

    def test_read_bands_duplicates(self):
        (i, j, k) = self.datum
        bands = (k - 5, k - 5, k, k -5)
        assert_almost_equal(self.image.read_bands(bands,
                                                  use_memmap=True)[i, j, 2],
                            self.value)
        assert_almost_equal(self.image.read_bands(bands,
                                                  use_memmap=False)[i, j, 2],
                            self.value)

    def test_read_subregion(self):
        (i, j, k) = self.datum
        region = self.image.read_subregion((i - 5, i + 9),
                                           (j - 3, j + 4), use_memmap=True)
        assert_almost_equal(region[5, 3, k], self.value)
        region = self.image.read_subregion((i - 5, i + 9),
                                           (j - 3, j + 4), use_memmap=False)
        assert_almost_equal(region[5, 3, k], self.value)

    def test_read_subimage(self):
        (i, j, k) = self.datum
        subimage = self.image.read_subimage([0, 3, i, 5],
                                            [1, j, 4, 7],
                                            [3, 7, k], use_memmap=True)
        assert_almost_equal(subimage[2, 1, 2], self.value)
        subimage = self.image.read_subimage([0, 3, i, 5],
                                            [1, j, 4, 7],
                                            [3, 7, k], use_memmap=False)
        assert_almost_equal(subimage[2, 1, 2], self.value)

        subimage = self.image.read_subimage([0, 3, i, 5],
                                            [1, j, 4, 7], use_memmap=True)
        assert_almost_equal(subimage[2, 1, k], self.value)
        subimage = self.image.read_subimage([0, 3, i, 5],
                                            [1, j, 4, 7], use_memmap=False)
        assert_almost_equal(subimage[2, 1, k], self.value)

    def test_load(self):
        (i, j, k) = self.datum
        data = self.image.load()
        spyf = self.image

        load_assert = np.allclose
        load_assert(data[i, j, k], self.value)
        first_band = spyf[:, :, 0]
        load_assert(data[:, :, 0], first_band)
        # This is checking if different ImageArray and SpyFile indexing
        # results are the same shape, so we can't just reuse the already
        # loaded first band.
        load_assert(data[:, 0, 0], spyf[:, 0, 0])
        load_assert(data[0, 0, 0], spyf[0, 0, 0])
        load_assert(data[0, 0], spyf[0, 0])
        load_assert(data[-1, -1, -1], spyf[-1, -1, -1])
        load_assert(data[-1, -3:-1], spyf[-1, -3:-1])
        load_assert(data[(6, 25)], spyf[(6, 25)])

        # The following test would currently fail, because
        # SpyFile.__get_item__ treats [6,25] the same as (6,25).

        # load_assert(data[[6, 25]],
        #             spyf[[6, 25]])

        load_assert(data.read_band(0), spyf.read_band(0))
        load_assert(data.read_bands([0, 1]), spyf.read_bands([0, 1]))
        load_assert(data.read_pixel(1, 2), spyf.read_pixel(1, 2))
        load_assert(data.read_subregion([0, 3], [1, 2]),
                    spyf.read_subregion([0, 3], [1, 2]))
        load_assert(data.read_subregion([0, 3], [1, 2], [0, 1]),
                    spyf.read_subregion([0, 3], [1, 2], [0, 1]))
        load_assert(data.read_subimage([0, 2, 4], [6, 3]),
                    spyf.read_subimage([0, 2, 4], [6, 3]))
        load_assert(data.read_subimage([0, 2], [6, 3], [0, 1]),
                    spyf.read_subimage([0, 2], [6, 3], [0, 1]))
        load_assert(data.read_datum(1,2,8), spyf.read_datum(1,2,8))

        ufunc_result = data + 1
        assert isinstance(ufunc_result, np.ndarray)
        assert not isinstance(ufunc_result, type(data))
        non_ufunc_result = data.diagonal()
        assert isinstance(non_ufunc_result, np.ndarray)
        assert not isinstance(non_ufunc_result, type(data))

    def test_getitem_i_j_k(self):
        (i, j, k) = self.datum
        assert_almost_equal(self.image[i, j, k], self.value)

    def test_getitem_i_j(self):
        (i, j, k) = self.datum
        assert_almost_equal(self.image[i, j][k], self.value)

    def test_getitem_i_j_kslice(self):
        (i, j, k) = self.datum
        assert_almost_equal(self.image[i, j, k-2:k+3:2][0, 0, 1], self.value)

    def test_getitem_islice_jslice(self):
        (i, j, k) = self.datum
        assert_almost_equal(self.image[i-3:i+3, j-3:j+3][3, 3, k], self.value)

def assert_same_shape_almost_equal(obj1, obj2, decimal=7, err_msg='',
                                   verbose=True):
    """
    Assert that two objects are almost equal and have the same shape.

    numpy.testing.assert_almost_equal does test for shape, but considers
    arrays with one element and a scalar to be the same.
    """
    # Types might be different since ImageArray stores things as
    # floats by default.
    if np.isscalar(obj1):
        assert np.isscalar(obj2), err_msg
    else:
        assert obj1.shape == obj2.shape, err_msg

    assert_almost_equal(obj1, obj2, decimal=decimal, err_msg=err_msg,
                        verbose=verbose)

class SpyFileTestSuite(object):
    '''Tests reading by byte orders, data types, and interleaves. For a
    specified image file name, the test suite will verify proper reading of
    data for various combinations of data type, interleave (BIL, BIP, BSQ),
    and  byte order (little- and big-endian). A new file is created
    for each combination of parameters for testing.
    '''
    def __init__(self, filename, datum, value, **kwargs):
        '''
        Arguments:

            `filename` (str):

                Name of the image file to be tested.

            `datum` (3-tuple of ints):

                (i, j, k) are the row, column and band of the datum  to be
                tested. 'i' and 'j' should be at least 10 pixels away from the
                edge of the associated image and `k` should have at least 10
                bands above and below it in the image.

            `value` (int or float):

                The scalar value associated with location (i, j, k) in
                the image.

        Keyword Arguments:

            `dtypes` (tuple of numpy dtypes):

                The file will be tested for all of the dtypes given. If not
                specified, only float32 an float64 will be tested.
        '''
        self.filename = filename
        self.datum = datum
        self.value = value
        self.dtypes = kwargs.get('dtypes', ('f4', 'f8'))
        self.dtypes = [np.dtype(d).name for d in self.dtypes]

    def run(self):
        print('\n' + '-' * 72)
        print('Running SpyFile read tests.')
        print('-' * 72)

        if not os.path.isdir(testdir):
            os.mkdir(testdir)
        image = spy.open_image(self.filename)
        basename = os.path.join(testdir,
                                os.path.splitext(self.filename)[0])
        interleaves = ('bil', 'bip', 'bsq')
        ends = ('big', 'little')
        cases = itertools.product(interleaves, self.dtypes, ends)
        for (inter, dtype, endian) in cases:
            fname = '%s_%s_%s_%s.hdr' % (basename, inter, dtype,
                                         endian)
            spy.envi.save_image(fname, image, interleave=inter,
                                dtype=dtype, byteorder=endian)
            msg = 'Running SpyFile read tests on %s %s %s-endian file ' \
                % (inter.upper(), np.dtype(dtype).name, endian)
            testimg = spy.open_image(fname)
            if testimg.using_memmap is True:
                print('\n' + '-' * 72)
                print(msg + 'using memmap...')
                print('-' * 72)
                test = SpyFileTest(testimg, self.datum, self.value)
                test.run()
                print('\n' + '-' * 72)
                print(msg + 'without memmap...')
                print('-' * 72)
                testimg._disable_memmap()
                test = SpyFileTest(testimg, self.datum, self.value)
                test.run()
            else:
                print('\n' + '-' * 72)
                print(msg + 'without memmap...')
                print('-' * 72)
                test = SpyFileTest(testimg, self.datum, self.value)
                test.run()


def run():
    tests = [('92AV3C.lan', (99, 99, 99), 2057.0)]
#    tests = [('92AV3C.lan', (99, 99, 99), 2057.0),
#             ('f970619t01p02_r02_sc04.a.rfl', (99, 99, 99), 0.2311),
#             ('cup95eff.int.hdr', (99, 99, 33), 0.1842)]
    for (fname, datum, value) in tests:
        try:
            check = find_file_path(fname)
            suite = SpyFileTestSuite(fname, datum, value,
                                     dtypes=('i2', 'i4', 'f4', 'f8'))
            suite.run()
        except FileNotFoundError:
            print('File "%s" not found. Skipping.' % fname)

if __name__ == '__main__':
    run()



'''
Package containing unit test modules for various functionality.

To run all unit tests, type the following from the system command line:

    # python -m spectral.tests.run
'''

from __future__ import absolute_import, division, print_function, unicode_literals

# If abort_on_fail is True, an AssertionError will be raised when a unit test
# fails; otherwise, the failure will be printed to stdout and testing will
# continue.
abort_on_fail = True

# Summary stats of unit test execution
_num_tests_run = 0
_num_tests_failed = 0

# Subdirectory to be created for unit test files
testdir = 'spectral_test_files'

from . import database
from . import spyfile
from . import transforms
from . import memmap
from . import envi
from . import spymath
from . import detectors
from . import classifiers
from . import dimensionality
from . import spatial
from . import iterators
from . import continuum

# List of all submodules to be run from the `run` submodule.
all_tests = [spyfile, memmap, iterators, transforms, envi, spymath, detectors,
             classifiers, dimensionality, spatial, database, continuum]



'''
Functions over spatial regions of images.
'''

from __future__ import absolute_import, division, print_function, unicode_literals

__all__ = ['map_window', 'map_outer_window_stats', 'map_class_ids',
           'map_classes']

import itertools
import numpy as np

import spectral as spy
from .algorithms import GaussianStats, iterator_ij

def get_window_bounds(nrows, ncols, height, width, i, j):
    '''Returns boundaries of an image window centered on a specified pixel.

    Arguments:

        `nrows` (int):

            Total number of rows in the image

        `ncols` (int):

            Total number of columns in the image

        `height` (int):

            Height of the desired window (in pixels)

        `width` (int):

            Width of the desired window (in pixels)

        `i` (int):

            Row index of the pixel

        `j` (int):

            Column index of the pixel

    Return value:

        A 4-tuple of ints of the form

            (row_start, row_stop, col_start, col_stop).
            
    The dimensions of the window will always be (`height`, `width`). For
    pixels near the border of the image where there are insufficient pixels
    between the specified pixel and image border, the window will be flush
    against the border of the image and the pixel position will be offset
    from the center of the widow.

    For an alternate function that clips window pixels near the border of the
    image, see `get_window_bounds_clipped`.
    '''
    if height > nrows or width > ncols:
        raise ValueError('Window size is too large for image dimensions.')

    rmin = i - height // 2
    rmax = rmin + height
    if rmin < 0:
        rmax = height
        rmin = 0
    elif rmax > nrows:
        rmax = nrows
        rmin = nrows - height

    cmin = j - width // 2
    cmax = cmin + width
    if cmin < 0:
        cmax = width
        cmin = 0
    elif cmax > ncols:
        cmax = ncols
        cmin = ncols - width

    return (rmin, rmax, cmin, cmax)

def get_window_bounds_clipped(nrows, ncols, height, width, i, j):
    '''Returns boundaries of an image window centered on a specified pixel.

    Arguments:

        `nrows` (int):

            Total number of rows in the image

        `ncols` (int):

            Total number of columns in the image

        `height` (int):

            Height of the desired window (in pixels)

        `width` (int):

            Width of the desired window (in pixels)

        `i` (int):

            Row index of the pixel

        `j` (int):

            Column index of the pixel

    Return value:

        A 4-tuple of ints of the form

            (row_start, row_stop, col_start, col_stop).
            
    Near the boder of the image where there are insufficient pixels between
    the specified pixel and the image border, the window will be clipped.

    For an alternate function that always returns a window with dimensions
    (`width`, `height`), see `get_window_bounds`.
    '''
    if height > nrows or width > ncols:
        raise ValueError('Window size is too large for image dimensions.')

    rmin = i - height // 2
    rmax = rmin + height
    if rmin < 0:
        rmin = 0
    elif rmax > nrows:
        rmax = nrows

    cmin = j - width // 2
    cmax = cmin + width
    if cmin < 0:
        cmin = 0
    elif cmax > ncols:
        cmax = ncols

    return (rmin, rmax, cmin, cmax)

def map_window(func, image, window, rslice=(None,), cslice=(None,),
               border='shift', dtype=None):
    '''Applies a function over a rolling spatial window.
    
    Arguments:

        `func` (callable):

            The function to apply. This function must accept two inputs:

            `X` (ndarray):

                The image data corresponding to the spatial window for the
                current pixel being evaluated. `X` will have shape
                `window + (N,)`, where `N` is the number of bands in the image.
                For pixels near the border of the image, the first two
                dimensions of `X` may be smaller if `border` is set to "clip".

            `ij` (2-tuple of integers):

                Indicates the row/column of the current pixel within the
                window. For `window` with even dimensions or for pixels near
                the image border, this may not correspond to the center pixel
                in the window.
    
        `image` (`SpyFile` or np.ndarray):

            The image on which the apply `func` with the specified window.

        `window` (int or 2-tuple of ints):

            The size of the window, in pixels. If this value is an integer,
            the height and width of the window will both be set to the value.
            Otherwise, `window` should be a tuple of the form (height, width).
            
        `rslice` (tuple):

            Tuple of `slice` parameters specifying at which rows the function
            should be applied. If not provided, `func` is applied to all rows.

        `cslice` (tuple):

            Tuple of `slice` parameters specifying at which columns the
            function should be applied. If not provided, `func` is applied to
            all columns.

        `border` (string, default "shift"):

            Indicates how to handles windows near the edge of the window. If
            the value is "shift", the window dimensions will alway be
            `(width, height)` but near the image border the pixel being
            iterated will be offset from the center of the window. If set to
            "clip", window regions falling outside the image border will be
            clipped and the window dimension will be reduced.

        `dtype` (np.dtype):

            Optional dtype for the output.

    Return value:

        Returns an np.ndarray with shape corresponding to the row and column
        start/stop indices and shape of `func` output.

    Examples:
    ---------

    To produce a new image that is a 3x3 pixel average of the input image:

    >>> f = lambda X, ij: np.mean(X.reshape((-1, X.shape[-1])), axis=0)
    >>> image_3x3 = map_window(f, image, 3)

    Perform a 5x5 pixel average but only retain values at every fifth row and
    column (i.e., simulate an image at one fifth resolution):

    >>> image.shape
    (145, 145, 220)
    >>> image_5x5 = map_window(f, image, 5, (2, -2, 5), (2, -2, 5))
    >>> image_5x5.shape
    (29, 29, 220)
    '''
    if isinstance(window, (list, tuple)):
        (height, width) = window[:]
    else:
        (height, width) = (window, window)

    if border == 'shift':
        get_window = get_window_bounds
    elif border == 'clip':
        get_window = get_window_bounds_clipped
    else:
        raise ValueError('Unrecognized border option.')

    (nrows, ncols) = image.shape[:2]

    # Row/Col indices at which to apply the windowed function
    rvals = list(range(*slice(*rslice).indices(nrows)))
    cvals = list(range(*slice(*cslice).indices(ncols)))

    def get_val(i, j):
        (r0, r1, c0, c1) = get_window(nrows, ncols, height, width, i, j)
        return func(image[r0:r1, c0:c1],
                    (i - r0, j - c0)).astype(dtype)

    return np.array([[get_val(r, c) for c in cvals]
                     for r in rvals]).astype(dtype)

def map_outer_window_stats(func, image, inner, outer, dim_out=1, cov=None,
                           dtype=None, rslice=(None,), cslice=(None,)):
    '''Maps a function accepting `GaussianStats` over a rolling spatial window.
    
    Arguments:

        `func` (callable):

            A callable object that will be applied to each pixel when the
            __call__ method is called for this object. The __call__ method
            of `func` must accept two arguments:

                - `X` (`GaussianStats`):

                    The Gaussian statistics computed from pixels in the outer
                    window (excluding the inner window).

                - `v` (ndarray):

                    An ndarray representing the pixel for which the window
                    was produced.

        `image` (`SpyFile` or np.ndarray):

            The image on which the apply `func` with the specified window.

        `inner` (int or 2-tuple of ints):

            The size of the inner window, in pixels. If this value is an integer,
            the height and width of the window will both be set to the given value.
            Otherwise, `inner` should be a tuple of the form (height, width).
            All pixels within the inner window are excluded from statistics
            computed for the associated pixel.
            
        `outer` (int or 2-tuple of ints):

            The size of the outer window, in pixels. If this value is an integer,
            the height and width of the window will both be set to the given value.
            Otherwise, `outer` should be a tuple of the form (height, width).
            All pixels in the outer window (but not in the inner window) are
            used to compute statistics for the associated pixel.
            
        `rslice` (tuple):

            Tuple of `slice` parameters specifying at which rows the function
            should be applied. If not provided, `func` is applied to all rows.

        `cslice` (tuple):

            Tuple of `slice` parameters specifying at which columns the
            function should be applied. If not provided, `func` is applied to
            all columns.

        `dtype` (np.dtype):

            Optional dtype for the output.

    Return value:

        Returns an np.ndarray whose elements are the result of mapping `func`
        to the pixels and associated window stats.

    Examples:
    ---------

    To create an RX anomaly detector with a 3x3 pixel inner window and 17x17
    outer window (note that `spectral.rx` already does this):

    >>> def mahalanobis(bg, x):
    ...     return (x - bg.mean).dot(bg.inv_cov).dot(x - bg.mean)
    ...
    >>> rx_scores = map_outer_window_stats(mahalanobis, image, 3, 17)

    '''
    mapper = WindowedGaussianBackgroundMapper(inner, outer, func, cov, dim_out,
                                              dtype)
    return mapper(image, rslice, cslice)

class WindowedGaussianBackgroundMapper(object):
    '''A class for procucing window statistics with an inner exclusion window.
    '''
    def __init__(self, inner, outer, function=None, cov=None, dim_out=None,
                 dtype=None):
        '''Creates a detector with the given inner/outer window.

        Arguments:

            `inner` (integer or 2-tuple of integers):

                Width and heigth of inner window, in pixels.

            `outer` (integer or 2-tuple of integers):

                Width and heigth of outer window, in pixels. Dimensions must
                be greater than inner window

            `function` (callable object):

                A callable object that will be applied to each pixel when the
                __call__ method is called for this object. The __call__ method
                of `function` must accept two arguments:

                    - A `GaussianStats` object.

                    - An ndarray representing the pixel for which the
                      were computed.

            `cov` (ndarray):

                An optional covariance to use. If this parameter is given,
                `cov` will be used for all RX calculations (background
                covariance will not be recomputed in each window). Only the
                background mean will be recomputed in each window).

            `dim_out` (int):

                The dimensionality of the output of `function` when called on
                a pixel spectrum. If this value is not specified, `function`
                will be checked to see if it has a `dim_out` member. If it
                does not, `dim_out` will be assumed to be 1.

            `dtype`:

                Optional dtype for the output array. If not specified,
                np.float32 is used.
        '''
        if isinstance(inner, (list, tuple)):
            self.inner = inner[:]
        else:
            self.inner = (inner, inner)
        if isinstance(outer, (list, tuple)):
            self.outer = outer[:]
        else:
            self.outer = (outer, outer)
        self.callable = function
        self.cov = cov
        self.dim_out = dim_out
        self.create_mask = None
        if dtype is not None:
            self.dtype = dtype
        else:
            self.dtype = np.float32

    def __call__(self, image, rslice=(None,), cslice=(None,)):
        '''Applies the objects callable function to the image data.

        Arguments:

            `image` (numpy.ndarray):

                An image with shape (R, C, B).

            `rslice` (tuple):

                Tuple of `slice` parameters specifying at which rows the function
                should be applied. If not provided, `func` is applied to all rows.

            `cslice` (tuple):

                Tuple of `slice` parameters specifying at which columns the
                function should be applied. If not provided, `func` is applied to
                all columns.

        Returns numpy.ndarray:

            An array whose elements correspond to the outputs from the
            object's callable function.
        '''
        (R, C, B) = image.shape
        (row_border, col_border) = [x // 2 for x in self.outer]

        if self.dim_out is not None:
            dim_out = self.dim_out
        elif hasattr(self.callable, 'dim_out') and \
          self.callable.dim_out is not None:
            dim_out = self.callable.dim_out
        else:
            dim_out = 1

        # Row/Col indices at which to apply the windowed function
        rvals = list(range(*slice(*rslice).indices(R)))
        cvals = list(range(*slice(*cslice).indices(C)))

        nrows_out = len(rvals)
        ncols_out = len(cvals)

        if dim_out > 1:
            x = np.ones((nrows_out, ncols_out, dim_out),
                        dtype=np.float32) * -1.0
        else:
            x = np.ones((nrows_out, ncols_out), dtype=self.dtype) * -1.0

        npixels = self.outer[0] * self.outer[1] - self.inner[0] * self.inner[1]
        if self.cov is None and npixels < B:
            raise ValueError('Window size provides too few samples for ' \
                             'image data dimensionality.')

        if self.create_mask is not None:
            create_mask = self.create_mask
        else:
            create_mask = inner_outer_window_mask_creator(image.shape,
                                                          self.inner,
                                                          self.outer)

        interior_mask = create_mask(R // 2, C // 2, True)[2].ravel()
        interior_indices = np.argwhere(interior_mask == 0).squeeze()

        (i_interior_start, i_interior_stop) = (row_border, R - row_border)
        (j_interior_start, j_interior_stop) = (col_border, C - col_border)

        status = spy._status
        status.display_percentage('Processing image: ')
        if self.cov is not None:
            # Since we already have the covariance, just use np.mean to get
            # means of the inner window and outer (including the inner), then
            # use those to calculate the mean of the outer window alone.
            background = GaussianStats(cov=self.cov)
            for i in range(nrows_out):
                for j in range(ncols_out):
                    (inner, outer) = create_mask(rvals[i], cvals[j], False)
                    N_in = (inner[1] - inner[0]) * (inner[3] - inner[2])
                    N_tot = (outer[1] - outer[0]) * (outer[3] - outer[2])
                    mean_out = np.mean(image[outer[0]: outer[1],
                                             outer[2]: outer[3]].reshape(-1, B),
                                             axis=0)
                    mean_in = np.mean(image[outer[0]: outer[1],
                                            outer[2]: outer[3]].reshape(-1, B),
                                            axis=0)
                    mean = mean_out * (float(N_tot) / (N_tot - N_in)) - \
                           mean_in * (float(N_in) / (N_tot - N_in))
                    background.mean = mean
                    x[i, j] = self.callable(background,
                                            image[rvals[i], cvals[j]])
                if i % (nrows_out // 10) == 0:
                    status.update_percentage(100. * i // nrows_out)
        else:
            # Need to calculate both the mean and covariance for the outer
            # window (without the inner).
            (h, w) = self.outer[:]
            for i in range(nrows_out):
                ii = rvals[i] - h // 2
                for j in range(ncols_out):
                    jj = cvals[j] - w // 2
                    if i_interior_start <= rvals[i] < i_interior_stop and \
                        j_interior_start <= cvals[j] < j_interior_stop:
                        X = image[ii : ii + h, jj : jj + w, :]
                        indices = interior_indices
                    else:
                        (inner, (i0, i1, j0, j1), mask) = \
                          create_mask(rvals[i], cvals[j], True)
                        indices = np.argwhere(mask.ravel() == 0).squeeze()
                        X = image[i0 : i1, j0 : j1, :]
                    X = np.take(X.reshape((-1, B)), indices, axis=0)
                    mean = np.mean(X, axis=0)
                    cov = np.cov(X, rowvar=False)
                    background = GaussianStats(mean, cov)
                    x[i, j] = self.callable(background,
                                            image[rvals[i], cvals[j]])
                if i % (nrows_out // 10) == 0:
                    status.update_percentage(100. * i / nrows_out)

        status.end_percentage()
        return x

def inner_outer_window_mask_creator(image_shape, inner, outer):
    '''Returns a function to give  inner/outer windows.

    Arguments:

        `image_shape` (tuple of integers):

            Specifies the dimensions of the image for which windows are to be
            produced. Only the first two dimensions (rows, columns) is used.

        `inner` (int or 2-tuple of integers):

            Height and width of the inner window, in pixels.

        `outer` (int or 2-tuple of integers):

            Height and width of the outer window, in pixels.

    Return value:

        A function that accepts the following arguments:

            `i` (int):

                Row index of pixel for which to generate the mask
                
            `j` (int):

                Row index of pixel for which to generate the mask

            `gen_mask` (bool, default False):

                A boolean flag indicating whether to return a boolean mask of
                shape (window[1], window[1]), indicating which pixels in the
                window should be used for background statistics calculations.

        If `gen_mask` is False, the return value is a 2-tuple of 4-tuples,
        where the 2-tuples specify the start/stop row/col indices for the
        inner and outer windows, respectively. Each of the 4-tuples is of the
        form (row_start, row_stop, col_start, col_stop).

        If `gen_mask` is True, a third element is added the tuple, which is
        the boolean mask for the inner/outer window.
    '''
    (R, C) = image_shape[:2]
    if isinstance(inner, (list, tuple)):
        (hi, wi) = inner[:]
    else:
        (hi, wi) = (inner, inner)
    if isinstance(outer, (list, tuple)):
        (ho, wo) = outer[:]
    else:
        (ho, wo) = (outer, outer)

    if wi > wo or hi > ho:
        raise ValueError('Inner window dimensions must be smaller than outer.')
    
    (ai, bi) = (hi // 2, wi // 2)
    (ao, bo) = (ho // 2, wo // 2)

    def create_mask(i, j, gen_mask=False):

        # Inner window
        inner_imin = i - ai
        inner_imax = inner_imin + hi
        if inner_imin < 0:
            inner_imax = hi
            inner_imin = 0
        elif inner_imax > R:
            inner_imax = R
            inner_imin = R - hi
        
        inner_jmin = j - bi
        inner_jmax = inner_jmin + wi
        if inner_jmin < 0:
            inner_jmax = wi
            inner_jmin = 0
        elif inner_jmax > C:
            inner_jmax = C
            inner_jmin = C - wi
        
        # Outer window
        outer_imin = i - ao
        outer_imax = outer_imin + ho
        if outer_imin < 0:
            outer_imax = ho
            outer_imin = 0
        elif outer_imax > R:
            outer_imax = R
            outer_imin = R - ho
        
        outer_jmin = j - bo
        outer_jmax = outer_jmin + wo
        if outer_jmin < 0:
            outer_jmax = wo
            outer_jmin = 0
        elif outer_jmax > C:
            outer_jmax = C
            outer_jmin = C - wo
        
        inner = (inner_imin, inner_imax, inner_jmin, inner_jmax)
        outer = (outer_imin, outer_imax, outer_jmin, outer_jmax)
        if not gen_mask:
            return (inner, outer)
        mask = np.zeros((ho, wo), dtype=np.bool)
        mask[inner_imin - outer_imin : inner_imax - outer_imin,
             inner_jmin - outer_jmin : inner_jmax - outer_jmin] = True
        return (inner, outer, mask)
    return create_mask

def map_class_ids(src_class_image, dest_class_image, unlabeled=None):
    '''Create a mapping between class labels in two classification images.

    Running a classification algorithm (particularly an unsupervised one)
    multiple times on the same image can yield similar results but with
    different class labels (indices) for the same classes. This function
    produces a mapping of class indices from one classification image to
    another by finding class indices that share the most pixels between the
    two classification images.

    Arguments:

        `src_class_image` (ndarray):

            An MxN integer array of class indices. The indices in this array
            will be mapped to indices in `dest_class_image`.
    
        `dest_class_image` (ndarray):

            An MxN integer array of class indices.

        `unlabeled` (int or array of ints):

            If this argument is provided, all pixels (in both images) will be
            ignored when counting coincident pixels to determine the mapping.
            If mapping a classification image to a ground truth image that has
            a labeled background value, set `unlabeled` to that value.

    Return Value:

        A dictionary whose keys are class indices from `src_class_image` and
        whose values are class indices from `dest_class_image`.

    .. seealso::

       :func:`map_classes`
    '''
    src_ids = list(set(src_class_image.ravel()))
    dest_ids = list(set(dest_class_image.ravel()))
    cmap = {}
    if unlabeled is not None:
        if isinstance(unlabeled, int):
            unlabeled = [unlabeled]
        for i in unlabeled:
            if i in src_ids:
                src_ids.remove(i)
                cmap[i] = i
            if i in dest_ids:
                dest_ids.remove(i)
    else:
        unlabeled = []
    N_src = len(src_ids)
    N_dest = len(dest_ids)

    # Create matrix of coincidence counts between classes in src and dest.
    matches = np.zeros((N_src, N_dest), np.uint16)
    for i in range(N_src):
        src_is_i = (src_class_image == src_ids[i])
        for j in range(N_dest):
            matches[i, j] = np.sum(np.logical_and(src_is_i,
                                                  dest_class_image == dest_ids[j]))

    unmapped = set(src_ids)
    dest_available = set(dest_ids)
    while len(unmapped) > 0:
        (i, j) = tuple(np.argwhere(matches == np.max(matches))[0])
        mmax = matches[i, j]
        if mmax == 0:
            # Nothing left to map. Pick unused indices from dest_class_image
            for (old, new) in zip(sorted(unmapped), sorted(dest_available)):
                cmap[old] = new
                unmapped.remove(old)
                dest_available.remove(new)
            for old in unmapped:
                # The list of target classes has been exhausted. Pick the
                # smallest dest value that isn't already used.
                def next_id():
                    for ii in itertools.count():
                        if ii not in unlabeled and ii not in cmap.values():
                            return ii
                cmap[old] = next_id()
            break
        cmap[src_ids[i]] = dest_ids[j]
        unmapped.remove(src_ids[i])
        dest_available.remove(dest_ids[j])
        matches[i, :] = 0
        matches[:, j] = 0
    return cmap

def map_classes(class_image, class_id_map, allow_unmapped=False):
    '''Modifies class indices according to a class index mapping.

    Arguments:

        `class_image`: (ndarray):

            An MxN array of integer class indices.

        `class_id_map`: (dict):

            A dict whose keys are indices from `class_image` and whose values
            are new values for the corresponding indices. This value is
            usually the output of :func:`map_class_ids`.

        `allow_unmapped` (bool, default False):

            A flag indicating whether class indices can appear in `class_image`
            without a corresponding key in `class_id_map`. If this value is
            False and an index in the image is found without a mapping key,
            a :class:`ValueError` is raised. If True, the unmapped index will
            appear unmodified in the output image.

    Return Value:

        An integer-valued ndarray with same shape as `class_image`

    Example:

        >>> m = spy.map_class_ids(result, gt, unlabeled=0)
        >>> result_mapped = spy.map_classes(result, m)

    .. seealso::

       :func:`map_class_ids`
    '''
    if not allow_unmapped  \
      and not set(class_id_map.keys()).issuperset(set(class_image.ravel())):
        raise ValueError('`src` has class values with no mapping key')
    mapped = np.array(class_image)
    for (i, j) in class_id_map.items():
        mapped[class_image == i] = j
    return mapped

def expand_binary_mask_for_window(mask, height, width):
    '''Returns a new mask including window around each pixel in source mask.

    Arguments:

        `mask` (2D ndarray):

            An ndarray whose non-zero elements define a mask.

        `height` (int):

            Height of the window.

        `width` (int):

            Width of the window

    Returns a new mask of ones and zeros with same shape as `mask`. For each
    non-zero element in mask, the returned mask will contain a value of one
    for all pixels in the `height`x`width` window about the pixel and zeros
    elsewhere.
    '''
    m = np.zeros_like(mask)
    (mask_height, mask_width) = mask.shape
    for (i, j) in iterator_ij(mask):
        (r0, r1, c0, c1) = get_window_bounds_clipped(mask_height, mask_width,
                                                     height, width, i, j)
        m[r0:r1, c0:c1] = 1
    return m
